'''
Specialized scientific functions for biogeophysical variables and L4C model
processes.
'''

import numpy as np
from functools import partial
from scipy.ndimage import generic_filter
from scipy.linalg import solve_banded
from scipy.sparse import dia_matrix
from pyl4c import suppress_warnings
from pyl4c.data.fixtures import HDF_PATHS, BPLUT
from pyl4c.utils import get_pft_array, subset
from pyl4c.stats import ols, ols_variance, linear_constraint

def arrhenius(
        tsoil, beta0: float, beta1: float = 66.02, beta2: float = 227.13):
    r'''
    The Arrhenius equation for response of enzymes to (soil) temperature,
    constrained to lie on the closed interval [0, 1].

    $$
    f(T_{SOIL}) = \mathrm{exp}\left[\beta_0\left( \frac{1}{\beta_1} -
        \frac{1}{T_{SOIL} - \beta_2} \right) \right]
    $$

    Parameters
    ----------
    tsoil : numpy.ndarray
        Array of soil temperature in degrees K
    beta0 : float
        Coefficient for soil temperature (deg K)
    beta1 : float
        Coefficient for ... (deg K)
    beta2 : float
        Coefficient for ... (deg K)

    Returns
    -------
    numpy.ndarray
        Array of soil temperatures mapped through the Arrhenius function
    '''
    a = (1.0 / beta1)
    b = np.divide(1.0, np.subtract(tsoil, beta2))
    # This is the simple answer, but it takes on values >1
    y0 = np.exp(np.multiply(beta0, np.subtract(a, b)))
    # Constrain the output to the interval [0, 1]
    return np.where(y0 > 1, 1, np.where(y0 < 0, 0, y0))


def bias_correction_parameters(
        series, npoly: int = 1, cutoff: float = 1, var_cutoff: float = None,
        add_intercept: bool = True):
    '''
    Calculate the bias correction parameters for two overlapping time series,
    nominally the Nature Run and L4C Operational products, of a given
    variable using quantile mapping. For example, can correct the bias in
    Nature Run (2000-2017) against the L4C Ops record (2015-Present) by
    fitting bias correction parameters for the overlap period 2015-2017.
    Model can be specified:

        y = alpha + X beta_0 + X^2 beta_1 + ...

    NOTE: Because Nature Run and L4C Ops compare very well in some locations,
    a degree-1 polynomial (straight line) is fit first (regardless of npoly);
    if this solution produces corrections that are <1 gC m^-2, the degree-1
    solution is used. In some areas, there is a strong linear correspondence
    between most measurements but a small number have a super-linear
    relationship that is poorly fit by a degree-2 polynomial; in these cases
    (where model variance of the degree-2 fit is > var_cutoff), the degree-1
    solution is used. Forcing the line of best fit through the origin (with
    intercept=False) is also not recommended.

    Parameters
    ----------
    series : numpy.ndarray
        A (t x 2) NumPy array where t rows correspond to t time steps and
        each column is a product; the first column is the reference product
        or dependent variable in the linear bias correction.
    npoly : int
        Degree of the polynomial to use in bias correction (Default: 1)
    cutoff : float
        Cutoff for the degree-1 bias correction, in data units (e.g.,
        1 g C m-2 day-1); defaults to 1.0, i.e., the residual after correction
        must be greater than 1 g C m-2 day-1, which is the average impact of
        L4SM versus model-only observations. If this cutoff is exceeded, the
        degree-1 solution is returned.
    var_cutoff : float or None
        Cutoff in variance for higher-order solutions; if the residual model
        variance exceeds this threshold for the degree-N solution, then return
        the degree (N-1) solution (Default: None)
    add_intercept : bool
        True to add a the y-intercept term (Default: True)

    Returns
    -------
    numpy.ndarray
        A vector of length N + 1 where N is the degree of the polynomial
        fit requested
    '''
    def xmat(x, npoly):
        # Creates the design/ model matrix for a polynomial series
        # Add a column for each power of the requested polynomial series
        x = np.repeat(x.reshape((t, 1)), npoly, axis = 1)
        for i in range(1, npoly):
            # Calculate X^n for n up to N powers
            x[:,i] = np.power(x[:,0], npoly + 1)
        return x

    def fit(x, y, npoly):
        # Fits the model using OLS
        # If all of the Y values are NaN
        if np.all(np.isnan(y)): return np.ones((npoly + 1,)) * np.nan
        try:
            return ols(xmat(x, npoly), y, add_intercept)
        except np.linalg.linalg.LinAlgError:
            return np.ones((npoly + 1,)) * np.nan

    # Sort the input series from low -> high
    t = series.shape[0]
    y = np.sort(series[:,0])
    x = np.sort(series[:,1])

    # For some pixels, the time series has zero variance, and this can produce
    #   unstable OLS estimates (e.g., zero slope)
    if np.var(y) == 0 or np.var(x) == 0:
        # Return coefficients: (0, 1, 0, ..., 0)
        return np.hstack(((0, 1), list(0 for i in range(1, npoly))))

    if np.var(y) == 0 and np.var(x) == 0:
        # Intercept (mean) is the only necessary predictor
        return np.hstack(((1, 0), list(0 for i in range(1, npoly))))

    fit1 = np.hstack(
        (fit(x, y, npoly = 1), list(0 for i in range(1, npoly))))
    if npoly == 1:
        return fit1

    # First, try a degree-1 polynomial (straight-line) fit; if the bias
    #   correction slope is such that the correction is < 1 gC/m^-2,
    #   which is similar to the average impact of L4SM vs. model-only
    #   observations, then use the degree-1 fit parameters
    if x.mean() - (fit1[1] * x.mean()) < cutoff:
        return fit1

    # Second, starting with the simpler model, check if progressively more
    #   complicated models (up to a maximum of npoly) really do fit the data
    #   better; if not, or if the model variance is above a cutoff, use the
    #   next most-complicated model (last_model)
    last_model = fit1 # Starting with the simplest model...
    for p in range(2, npoly + 1):
        model = fit(x, y, npoly = p)
        # Calculates unbiased estimate of model variance
        model_var = ols_variance(xmat(x, p), y, model, add_intercept)
        # Without a cutoff for guidance, if the model variance of the degree-1
        #   fit is lower than that of the degree-2 fit...
        if var_cutoff is None:
            if model_var > ols_variance(
                    xmat(x, 1), y, last_model[0:p], add_intercept):
                return last_model
        else:
            if model_var > var_cutoff:
                return last_model

        last_model = model

    # Unless a simpler model was better, return coefficients for the requested
    #   polynomial degree
    return model


def climatology365(series, dates, ignore_leap = True):
    '''
    Computes a 365-day climatology for different locations from a time series
    of length T. The climatology could then be indexed using ordinals
    generated by `ordinals365()`. Setting `ignore_leap = False` may be useful
    if the time series has regular dropouts; e.g., for MODIS 8-day composite
    data, there are only 46 days each year with valid data.

    Parameters
    ----------
    series : numpy.ndarray
        T x ... array of data
    dates : list or tuple
        Sequence of datetime.datetime or datetime.date instances
    ignore_leap : bool
        True to convert DOY to (DOY-1) in leap years, effectively ignoring
        Leap Day (Default); if False, DOY numbers are unchanged

    Returns
    -------
    numpy.ndarray
    '''
    @suppress_warnings
    def calc_climatology(x):
        return np.array([
            np.nanmean(x[ordinal == day,...], axis = 0)
            for day in range(1, 366)
        ])
    # Get first and last day of the year (DOY)
    ordinal = np.array([
        # Finally, subtract 1 from each day in a leap year after Leap Day
        (doy - 1) if (
            ignore_leap and (dates[i].year % 4 == 0) and doy >= 60) else doy
        for i, doy in enumerate([
            # Next, fill in 0 wherever Leap Day occurs
            0 if (dates[i].year % 4 == 0 and doy == 60) else doy
            for i, doy in enumerate([
                # First, convert datetime.datetime to ordinal day-of-year (DOY)
                int(dt.strftime('%j')) for dt in dates
            ])
        ])
    ])
    return calc_climatology(series)


def daynight_partition(arr_24hr, updown, reducer = 'mean'):
    '''
    Partitions a 24-hour time series array into daytime and nighttime values,
    then calculates the mean in each group. Daytime is defined as when the sun
    is above the horizon; nighttime is the complement.

    Parameters
    ----------
    arr_24hr : numpy.ndarray
        A size (24 x ...) array; the first axis must have 24 elements
        corresponding to the measurement in each hour
    updown: numpy.ndarray
        A size (2 x ...) array, compatible with arr_24hr, where the first axis
        has the hour of sunrise and sunset, in that order, for each element
    reducer : str
        One of "mean" or "sum" indicating whether an average or a total of the
        daytime/ nighttime values should be calculated; e.g., for "mean", the
        hourly values from daytime hours are added up and divided by the
        length of the day (in hours).

    Returns
    -------
    numpy.ndarray
        A size (2 x ...) array where the first axis enumerates the daytime and
        nighttime mean values, respectively
    '''
    assert reducer in ('mean', 'sum'),\
        'Argument "reducer" must be one of: "mean", "sum"'
    # Prepare single-valued output array
    arr_daytime = np.zeros(arr_24hr.shape[1:])
    arr_nighttime = arr_daytime.copy()
    daylight_hrs = arr_daytime.copy().astype(np.int16)
    # Do sunrise and sunset define an interval? (Sunset > Sunrise)?
    inside_interval = np.apply_along_axis(lambda x: x[1] > x[0], 0, updown)
    # Or is the sun never up?
    never_up = np.logical_and(updown[0,...] == -1, updown[1,...] == -1)
    # Iteratively sum daytime VPD and temperature values
    for hr in range(0, 24):
        # Given only hour of sunrise/set on a 24-hour clock...
        #   if sun rises and sets on same day: SUNRISE <= HOUR <= SUNSET;
        #   if sun sets on next day: either SUNRISE <= HOUR or HOUR <= SUNSET;
        sun_is_up = np.logical_or( # Either...
            np.logical_and(inside_interval, # ...Rises and sets same day
                np.logical_and(updown[0,...] <= hr, hr <= updown[1,...])),
            np.logical_and(~inside_interval, # ...Sets on next day
                np.logical_or(updown[0,...] <= hr, hr <= updown[1,...])))
        # For simplicity, compute a 24-hour mean even if the sun never rises;
        #   there's no way to know what the "correct" daytime value is
        mask = np.logical_or(never_up, sun_is_up)
        np.add(np.where(
            mask, arr_24hr[hr,...], 0), arr_daytime, out = arr_daytime)
        np.add(np.where(
            ~mask, arr_24hr[hr,...], 0), arr_nighttime, out = arr_nighttime)
        # Keep track of the denominator (hours) for calculating the mean;
        #   note that this over-estimates actual daylight hours by 1 hour
        #   but results in the correct denominator for the sums above
        np.add(np.where(mask, 1, 0), daylight_hrs, out = daylight_hrs)
    arr_24hr = None
    # Calculate mean quantities
    if reducer == 'mean':
        arr_daytime = np.divide(arr_daytime, daylight_hrs)
        arr_nighttime = np.divide(arr_nighttime, 24 - daylight_hrs)
        # For sites where the sun is always above/ below the horizon, set missing
        #   nighttime values to zero
        arr_nighttime[~np.isfinite(arr_nighttime)] = 0
    return np.stack((arr_daytime, arr_nighttime))


def degree_lengths(phi, a = 6378137, b = 6356752.3142):
    '''
    Returns the approximate length of degrees of latitude and longitude.
    Source:

        https://en.wikipedia.org/wiki/Latitude

    phi : Number
        Latitude, in degrees
    a : Number
        Radius of the Earth (major axis) in meters
    b : Number
        Length of minor axis of the Earth in meters

    Returns
    -------
    tuple
        Length of a degree of (longitude, latitude), respectively
    '''
    e2 = ((a**2) - (b**2)) / (a**2)
    # Approximate length of a degree of latitude
    lat_m = 111132.954 - (559.822 * np.cos(2 * np.deg2rad(phi))) +\
        (1.175 * np.cos(4 * np.deg2rad(phi)))
    lng_m = (np.pi * a * np.cos(np.deg2rad(phi))) / (
        180 * np.sqrt(1 - (e2 * np.sin(np.deg2rad(phi))**2)))
    return (lng_m, lat_m)


def e_mult(params, tmin, vpd, smrz, ft):
    '''
    Calculate environmental constraint multiplier for gross primary
    productivity (GPP), E_mult, based on current model parameters. The
    expected parameter names are "LUE" for the maximum light-use
    efficiency; "smrz0" and "smrz1" for the lower and upper bounds on root-
    zone soil moisture; "vpd0" and "vpd1" for the lower and upper bounds on
    vapor pressure deficity (VPD); "tmin0" and "tmin1" for the lower and
    upper bounds on minimum temperature; and "ft0" for the multiplier during
    frozen ground conditions.

    Parameters
    ----------
    params : dict
        A dict-like data structure with named model parameters
    tmin : numpy.ndarray
        (T x N) vector of minimum air temperature (deg K), where T is the
        number of time steps, N the number of sites
    vpd : numpy.ndarray
        (T x N) vector of vapor pressure deficit (Pa), where T is the number
        of time steps, N the number of sites
    smrz : numpy.ndarray
        (T x N) vector of root-zone soil moisture wetness (%), where T is the
        number of time steps, N the number of sites
    ft : numpy.ndarray
        (T x N) vector of the (binary) freeze-thaw status, where T is the
        number of time steps, N the number of sites (Frozen = 0, Thawed = 1)

    Returns
    -------
    numpy.ndarray
    '''
    # Calculate E_mult based on current parameters
    f_tmin = linear_constraint(params['tmin0'], params['tmin1'])
    f_vpd  = linear_constraint(params['vpd0'], params['vpd1'], 'reversed')
    f_smrz = linear_constraint(params['smrz0'], params['smrz1'])
    f_ft   = linear_constraint(params['ft0'], 1.0, 'binary')
    return f_tmin(tmin) * f_vpd(vpd) * f_smrz(smrz) * f_ft(ft)


def k_mult(params, tsoil, smsf):
    '''
    Calculate environmental constraint multiplier for soil heterotrophic
    respiration (RH), K_mult, based on current model parameters. The expected
    parameter names are "tsoil" for the Arrhenius function of soil temperature
    and "smsf0" and "smsf1" for the lower and upper bounds of the ramp
    function on surface soil moisture.

    Parameters
    ----------
    params : dict
        A dict-like data structure with named model parameters
    tsoil : numpy.ndarray
        (T x N) vector of soil temperature (deg K), where T is the number of
        time steps, N the number of sites
    smsf : numpy.ndarray
        (T x N) vector of surface soil wetness (%), where T is the number of
        time steps, N the number of sites

    Returns
    -------
    numpy.ndarray
    '''
    f_tsoil = partial(arrhenius, beta0 = params['tsoil'])
    f_smsf  = linear_constraint(params['smsf0'], params['smsf1'])
    return f_tsoil(tsoil) * f_smsf(smsf)


def litterfall_casa(lai, years, dt = 1/365):
    '''
    Calculates daily litterfall fraction after the CASA model (Randerson et
    al. 1996). Computes the fraction of evergreen versus deciduous canopy and
    allocates a constant daily fraction (out of the year) for evergreen canopy
    but a varying daily fraction for deciduous, where the fraction varies with
    "leaf loss," a function of leaf area index (LAI). Canopies are assumed to
    be a mix of evergreen and deciduous, so the litterfall fraction is a sum
    of these two approaches.

    Randerson, J. T., Thompson, M. V, Malmstrom, C. M., Field, C. B., &
      Fung, I. Y. (1996). Substrate limitations for heterotrophs: Implications
      for models that estimate the seasonal cycle of atmospheric CO2.
      *Global Biogeochemical Cycles,* 10(4), 585–602.

    The approach here is a bit different from Randerson et al. (1996) because
    we re- calculate the evergreen fraction each year; however, this is a
    reasonable elaboration that, incidentally, accounts for potential changes
    in the evergreen-vs-deciduous mix of the canopy. The result is an array
    of daily litterfall fractions, i.e., the result multiplied by the annual
    NPP sum (for a given site and year) obtains the daily litterfall.

    Parameters
    ----------
    lai : numpy.ndarray
        The (T x N) leaf-area index (LAI) array, for T time steps and N sites
    years : numpy.ndarray
        A length-T 1D array indexing the years, e.g., [2001, 2001, 2001, ...];
        used to identify which of T time steps belong to a year, so that
        litterfall fractions sum to one over a year
    dt : float
        The fraction of a year that each time step represents, e.g., for daily
        time steps, should be close to 1/365 (Default: 1/365)

    Returns
    -------
    numpy.ndarray
        The fraction of available inputs (e.g., annual NPP) that should be
        allocated to litterfall at each time step
    '''
    def leaf_loss(lai):
        # Leaf loss function from CASA, a triangular averaging function
        #   centered on the current date, where the right limb of the
        #   triangle is subtracted from the left limb (leading minus
        #   lagged LAI is equated to leaf loss)
        ll = generic_filter(
            lai, lambda x: (0.5 * x[0] + x[1]) - (x[3] + 0.5 * x[4]),
            size = 5, mode = 'mirror')
        return np.where(ll < 0, 0, ll) # Leaf loss cannot be < 0

    # Get leaf loss at each site (column-wise)
    ll = np.apply_along_axis(leaf_loss, 0, lai)
    ll = np.where(np.isnan(ll), 0, ll) # Fill NaNs with zero leaf loss
    unique_years = np.unique(years).tolist()
    unique_years.sort()
    for each_year in unique_years:
        # For those dates in this year...
        idx = years == each_year
        # Calculate the evergreen fraction (ratio of min LAI to mean LAI over
        #   the course of a year)
        efrac = np.apply_along_axis(
            lambda x: np.nanmin(x) / np.nanmean(x), 0, lai[idx,:])
        # Calculate sum of 1/AnnualNPP (Evergreen input) plus daily leaf loss
        #   fraction (Deciduous input); Evergreen canopies have constant daily
        #   inputs
        ll[idx,:] = (efrac * dt) + (1 - efrac) * np.divide(
            ll[idx,:], ll[idx,:].sum(axis = 0))
    return ll


def mean_residence_time(
        hdf, units = 'years', subset_id = None, nodata = -9999):
    '''
    Calculates the mean residence time (MRT) of soil organic carbon (SOC)
    pools as the quotient of SOC stock size and heterotrophic respiration
    (RH). Chen et al. (2013, Global and Planetary Change), provide a formal
    equation for mean residence time: (SOC/R_H).

    Parameters
    ----------
    hdf : h5py.File
        The HDF5 file / h5py.File object
    units : str
        Either "years" (default) or "days"
    subset_id : str
        (Optional) Can provide keyword designating the desired subset area
    nodata : float
        (Optional) The NoData or Fill value (Default: -9999)

    Returns
    -------
    tuple
        Tuple of: subset array, xoff, yoff, i.e., (numpy.ndarray, Int, Int)
    '''
    assert units in ('days', 'years'), 'The units argument must be one of: "days" or "years"'
    soc_field = HDF_PATHS['SPL4CMDL']['4']['SOC']
    rh_field = HDF_PATHS['SPL4CMDL']['4']['RH']
    if subset_id is not None:
        # Get X- and Y-offsets while we're at it
        soc, xoff, yoff = subset(
            hdf, soc_path, None, None, subset_id = subset_id)
        rh, _, _ = subset(
            hdf, rh_path, None, None, subset_id = subset_id)
    else:
        xoff = yoff = 0
        soc = hdf[soc_path][:]
        rh = hdf[rh_path][:]

    # Find those areas of NoData in either array
    mask = np.logical_or(soc == nodata, rh == nodata)
    mrt = np.divide(soc, rh)
    if units == 'years':
        # NOTE: No need to guard against NaNs/ NoData here because of mask
        mrt = np.divide(mrt, 365.0)
    np.place(mrt, mask, nodata) # Put NoData values back in
    return (mrt, xoff, yoff)


def npp(
        hdf, use_subgrid = False, subset_id = None, subset_bbox = None,
        nodata = -9999):
    '''
    Calculates net primary productivity (NPP), based on the carbon use
    efficiency (CUE) of each plant functional type (PFT). NPP is derived
    as: `NPP = GPP * CUE`, where `CUE = NPP/GPP`.

    Parameters
    ----------
    hdf : h5py.File
        The HDF5 file / h5py.File object
    use_subgrid : bool
        True to use the 1-km subgrid; requires iterating through the PFT means
    subset_id : str
        (Optional) Can provide keyword designating the desired subset area
    subset_bbox : list or tuple
        (Optional) Can provide a bounding box to define a desired subset area
    nodata : float
        The NoData value to mask (Default: -9999)

    Returns
    -------
    numpy.ndarray
        NPP values on an EASE-Grid 2.0 array
    '''
    grid = 'M01' if use_subgrid else 'M09'
    cue_array = cue(get_pft_array(grid, subset_id, subset_bbox))
    if not use_subgrid:
        if subset_id is not None or subset_bbox is not None:
            gpp, _, _ = subset(
                hdf, 'GPP/gpp_mean', subset_id = subset_id,
                subset_bbox = subset_bbox)
        else:
            gpp = hdf['GPP/gpp_mean'][:]
    else:
        raise NotImplementedError('No support for the 1-km subgrid')
    gpp[gpp == nodata] = np.nan
    return np.multiply(gpp, cue_array)


def ordinals365(dates):
    '''
    Returns a length-T sequence of ordinals on [1,365]. Can be used for
    indexing a 365-day climatology; see `climatology365()`.

    Parameters
    ----------
    dates : list or tuple
        Sequence of datetime.datetime or datetime.date instances

    Returns
    -------
    list
    '''
    return [
        t - 1 if (year % 4 == 0 and t >= 60) else t
        for t, year in [(int(t.strftime('%j')), t.year) for t in dates]
    ]


def rescale_smrz(smrz0, smrz_min, smrz_max = 100):
    '''
    Rescales root-zone soil-moisture (SMRZ); original SMRZ is in percent
    saturation units. NOTE: Although Jones et al. (2017) write "SMRZ_wp is
    the plant wilting point moisture level determined by ancillary soil
    texture data provided by L4SM..." in actuality it is just `smrz_min`.

    Parameters
    ----------
    smrz0 : numpy.ndarray
        (T x N) array of original SMRZ data, in percent (%) saturation units
        for N sites and T time steps
    smrz_min : numpy.ndarray or float
        Site-level long-term minimum SMRZ (percent saturation)
    smrz_max : numpy.ndarray or float
        Site-level long-term maximum SMRZ (percent saturation); can optionally
        provide a fixed upper-limit on SMRZ; useful for calculating SMRZ100.

    Returns
    -------
    numpy.ndarray
    '''
    if smrz_min.ndim == 1:
        smrz_min = smrz_min[np.newaxis,:]
    assert smrz0.ndim == 2,\
        'Expected smrz0 to be a 2D array'
    assert smrz0.shape[1] == smrz_min.shape[1],\
        'smrz_min should have one value per site'
    # Clip input SMRZ to the lower, upper bounds
    smrz0 = np.where(smrz0 < smrz_min, smrz_min, smrz0)
    smrz0 = np.where(smrz0 > smrz_max, smrz_max, smrz0)
    smrz_norm = np.add(np.multiply(100, np.divide(
        np.subtract(smrz0, smrz_min),
        np.subtract(smrz_max, smrz_min))), 1)
    # Log-transform normalized data and rescale to range between
    #   5.0 and 100% saturation)
    return np.add(
        np.multiply(95, np.divide(np.log(smrz_norm), np.log(101))), 5)


def soc_analytical_spinup(litterfall, k_mult, fmet, fstr, decay_rates):
    r'''
    Using the solution to the differential equations governing change in the
    soil organic carbon (SOC) pools, calculates the steady-state size of each
    SOC pool.

    The analytical steady-state value for the metabolic ("fast") pool is:
    $$
    C_{met} = \frac{f_{met} \sum NPP}{R_{opt} \sum K_{mult}}
    $$

    The analytical steady-state value for the structural ("medium") pool is:
    $$
    C_{str} = \frac{(1 - f_{met})\sum NPP}{R_{opt}\, k_{str} \sum K_{mult}}
    $$

    The analytical steady-state value for the recalcitrant ("slow") pool is:
    $$
    C_{rec} = \frac{f_{str}\, k_{str}\, C_{str}}{k_{rec}}
    $$

    Parameters
    ----------
    litterfall : numpy.ndarray
        Average daily litterfall, a (N x ...) array
    k_mult : numpy.ndarray
        The K_mult climatology, i.e., a (365 x N x ...) array of the long-term
        average K_mult value at each of N sites (optionally, with 81 1-km
        subgrid sites, e.g., 365 x N x 81)
    fmet : numpy.ndarray
        The f_metabolic model parameter, as an (N x ...) array
    fstr : numpy.ndarray
        The f_structural model parameter, as an (N x ...) array
    decay_rates : numpy.ndarray
        The optimal decay rates for each SOC pool, as a (3 x N x ...) array

    NOTE: If a 3 or more axes are used, those axes must match for all arrays;
    i.e., if (x ...) is used, it must be the same for all.

    Returns
    -------
    tuple
        A 3-element tuple, each element the steady-state values for that pool,
        i.e., `(metabolic, structural, recalcitrant)`
    '''
    # NOTE: litterfall is average daily litterfall (see upstream where we
    #   divided by 365), so, to obtain annual sum, multiply by 365
    c0 = np.divide(
        fmet * (litterfall * 365),
        decay_rates[0,...] * np.sum(k_mult, axis = 0))
    c1 = np.divide(
        (1 - fmet) * (litterfall * 365),
        decay_rates[1,...] * np.sum(k_mult, axis = 0))
    # NOTE: k_mult disappears because it is in both the numerator and
    #   denominator
    c2 = np.divide(fstr * decay_rates[1,...] * c1, decay_rates[2,...])
    c0[~np.isfinite(c0)] = 0
    c1[~np.isfinite(c1)] = 0
    c2[~np.isfinite(c2)] = 0
    return (c0, c1, c2)


def soc_numerical_spinup(
        soc, litterfall, k_mult, fmet, fstr, decay_rates, threshold = 0.1,
        verbose = False):
    '''
    Numerical spin-up of C pools.

    Parameters
    ----------
    soc : numpy.ndarray
        SOC in each pool in g C m-3 units, a (3 x N x ...) array
    litterfall : numpy.ndarray
        Daily litterfall in g C m-2 units, a (N x ...) array
    k_mult : numpy.ndarray
        The K_mult climatology, i.e., a (365 x N x ...) array of the long-term
        average K_mult value at each of N sites (with ... 1-km subgrid sites)
    fmet : numpy.ndarray
        The f_metabolic model parameter, as an (N x ...) array
    fstr : numpy.ndarray
        The f_structural model parameter, as an (N x ...) array
    decay_rates : numpy.ndarray
        The optimal decay rates for each SOC pool, as a (3 x N x ...) array
    threshold : float
        Goal for the NEE tolerance check; i.e., change in NEE between
        climatological years should be less than or equal to the threshold
        for all pixels (Default: 0.1 g C m-2 yr-1)
    verbose : bool
        True to print messages to the screen

    Returns
    -------
    tuple
        2-element tuple of `((soc0, soc1, soc2), tol)` where the first
        element is a 3-tuple of the SOC in each pool; second element is the
        final tolerance
    '''
    tsize = k_mult.shape[0] # Whether 365 (days) or T days
    tol = np.inf
    i = 0
    # Jones et al. (2017) write that goal is NEE tolerance <=
    #   1 g C m-2 year-1, but we can do better
    if verbose:
        print('Iterating...')
    while not np.all(abs(tol) <= threshold):
        diffs = np.zeros(k_mult.shape)
        for t in range(0, tsize):
            rh = k_mult[t,...] * decay_rates * soc
            # Calculate change in C pools (g C m-2 units)
            dc0 = np.subtract(np.multiply(litterfall, fmet), rh[0])
            dc1 = np.subtract(np.multiply(litterfall, 1 - fmet), rh[1])
            dc2 = np.subtract(np.multiply(fstr, rh[1]), rh[2])
            soc[0] += dc0
            soc[1] += dc1
            soc[2] += dc2
            diffs[t,:] += (dc0 + dc1 + dc2)
        # Calculate total annual change in NEE ("mean tolerance") at each
        #   site over the year
        if i > 0:
            # Tolerance goes to zero as each successive year brings fewer
            #   changes in NEE
            tol = last_year - np.nansum(diffs, axis = 0)
        last_year = np.nansum(diffs, axis = 0)
        tol = np.where(np.isnan(tol), 0, tol)
        # Calculate mean absolute tolerance across sites
        if verbose:
            print('[%d] Mean (Max) abs. tolerance: %.4f (%.4f)' % (
                i, np.abs(tol).mean(), np.abs(tol).max()))
        i += 1
    return ((soc[0], soc[1], soc[2]), tol)


def soc_numerical_spinup2(
        soc, litterfall, k_mult, fmet, fstr, decay_rates, cue,
        threshold = 0.1, verbose = False):
    '''
    Numerical spin-up of C pools; here, the "tolerance" of spin-up is equal
    the annual NEE sum.

    Parameters
    ----------
    soc : numpy.ndarray
        SOC in each pool in g C m-3 units, a (3 x N x ...) array
    litterfall : numpy.ndarray
        Daily litterfall in g C m-2 units, a (N x ...) array
    k_mult : numpy.ndarray
        The K_mult climatology, i.e., a (365 x N x ...) array of the long-term
        average K_mult value at each of N sites (with ... 1-km subgrid sites)
    fmet : numpy.ndarray
        The f_metabolic model parameter, as an (N x ...) array
    fstr : numpy.ndarray
        The f_structural model parameter, as an (N x ...) array
    decay_rates : numpy.ndarray
        The optimal decay rates for each SOC pool, as a (3 x N x ...) array
    cue : numpy.ndarray
        The carbon use efficiency (CUE), a (N x ...) array
    threshold : float
        Goal for the NEE tolerance check; i.e., change in NEE between
        climatological years should be less than or equal to the threshold
        for all pixels (Default: 0.1 g C m-2 yr-1)
    verbose : bool
        True to print messages to the screen

    Returns
    -------
    tuple
        2-element tuple of `((soc0, soc1, soc2), tol)` where the first
        element is a 3-tuple of the SOC in each pool; second element is the
        final tolerance
    '''
    tsize = k_mult.shape[0] # Whether 365 (days) or T days
    tol = np.inf
    i = 0
    # Jones et al. (2017) write that goal is NEE tolerance <=
    #   1 g C m-2 year-1, but we can do better
    if verbose:
        print('Iterating...')
    while not np.all(abs(tol) <= threshold):
        nee = np.zeros(k_mult.shape)
        for t in range(0, tsize):
            rh = k_mult[t] * decay_rates[0] * soc
            # Calculate change in C pools (g C m-2 units)
            dc0 = np.subtract(np.multiply(litterfall, fmet), rh[0])
            dc1 = np.subtract(np.multiply(litterfall, 1 - fmet), rh[1])
            dc2 = np.subtract(np.multiply(fstr, rh[1]), rh[2])
            soc[0] += dc0
            soc[1] += dc1
            soc[2] += dc2
            # Adjust structural RH pool for material transferred to recalcitrant
            rh[1] = rh[2] * (1 - fstr)
            # Compute (mean daily) GPP as the (mean daily NPP):CUE ratio, then
            #   compute RA as (GPP - NPP)
            gpp = litterfall / cue
            # While it looks like we can optimize above+below, we'll need to
            #   re-use "gpp" later to calculate NEE ("diffs")
            ra = gpp - litterfall
            nee[t] = (ra + rh.sum(axis = 0)) - gpp
        if i > 0:
            # Tolerance goes to zero as each successive year brings fewer
            #   changes in NEE
            tol = last_year - np.nansum(nee, axis = 0)
        last_year = np.nansum(nee, axis = 0)
        tol = np.where(np.isnan(tol), 0, tol)
        # Calculate mean absolute tolerance across sites
        if verbose:
            print('[%d] Mean (Max) abs. tolerance: %.4f (%.4f)' % (
                i, np.abs(tol).mean(), np.abs(tol).max()))
        i += 1
    return ((soc[0], soc[1], soc[2]), tol)


def tridiag_solver(tri, r, kl = 1, ku = 1, banded = None):
    '''
    Solution to the tridiagonal equation by solving the system of equations
    in sparse form. Creates a banded matrix consisting of the diagonals,
    starting with the lowest diagonal and moving up, e.g., for matrix:

        A = [[10.,  2.,  0.,  0.],
             [ 3., 10.,  4.,  0.],
             [ 0.,  1.,  7.,  5.],
             [ 0.,  0.,  3.,  4.]]
        banded = [[ 3.,  1.,  3.,  0.],
                  [10., 10.,  7.,  4.],
                  [ 0.,  2.,  4.,  5.]]

    The banded matrix is what should be provided to the optoinal "banded"
    argument, which should be used if the banded matrix can be created faster
    than `scipy.sparse.dia_matrix()`.

    Parameters
    ----------
    tri : numpy.ndarray
        A tridiagonal matrix (N x N)
    r : numpy.ndarray
        Vector of solutions to the system, Ax = r, where A is the tridiagonal
        matrix
    kl : int
        Lower bandwidth (number of lower diagonals) (Default: 1)
    ku : int
        Upper bandwidth (number of upper diagonals) (Default: 1)
    banded : numpy.ndarray
        (Optional) Provide the banded matrix with diagonals along the rows;
        this can be faster than scipy.sparse.dia_matrix()

    Returns
    -------
    numpy.ndarray
    '''
    assert tri.ndim == 2 and (tri.shape[0] == tri.shape[1]),\
        'Only supports 2-dimensional square matrices'
    if banded is None:
        banded = dia_matrix(tri).data
    # If it is necessary, in a future implementation, to extract diagonals;
    #   this is a starting point for problems where kl = ku = 1
    # n = tri.shape[0]
    # a, b, c = [ # (n-1, n, n-1) refer to the lengths of each vector
    #     sparse[(i+1),(max(0,i)):j]
    #     for i, j in zip(range(-1, 2), (n-1, n, n+1))
    # ]
    return solve_banded((kl, ku), np.flipud(banded), r)


def vpd(qv2m, ps, temp_k):
	r'''
    Calculates vapor pressure deficit (VPD); unfortunately, the provenance
    of this formula cannot be properly attributed. It is taken from the
    SMAP L4C Science code base, so it is exactly how L4C calculates VPD.

    $$
    \mathrm{VPD} = 610.7 \times \mathrm{exp}\left(
    \frac{17.38 \times T_C}{239 + T_C}
    \right) - \frac{(P \times [\mathrm{QV2M}]}{0.622 + (0.378 \times [\mathrm{QV2M}])}
    $$

    Where P is the surface pressure (Pa), QV2M is the water vapor mixing
    ratio at 2-meter height, and T is the temperature in degrees C (though
    this function requires units of Kelvin when called).

    NOTE: A variation on this formula can be found in the text:

    Monteith, J. L. and M. H. Unsworth. 1990.
    Principles of Environmental Physics, 2nd. Ed. Edward Arnold Publisher.

    See also:
        https://glossary.ametsoc.org/wiki/Mixing_ratio

    Parameters
    ----------
    qv2m : numpy.ndarray or float
        QV2M, the water vapor mixing ratio at 2-m height
    ps : numpy.ndarray or float
        The surface pressure, in Pascals
    temp_k : numpy.ndarray or float
        The temperature at 2-m height in degrees Kelvin

    Returns
    -------
    numpy.ndarray or float
        VPD in Pascals
    '''
	temp_c = temp_k - 273.15 # Convert temperature to degrees C
	avp = np.divide(np.multiply(qv2m, ps), 0.622 + (0.378 * qv2m))
	x = np.divide(17.38 * temp_c, (239 + temp_c))
	esat = 610.7 * np.exp(x)
	return np.subtract(esat, avp)



'''
For all things related to TransCom regions but, in particular, for
the statistical summary of SMAP L4C (or other raster array data) by TransCom
region. The TransCom project seems to be poorly documented, but Carbon
Tracker [1] uses it and provides the data used here. Intended use (Example):

    >>> f = h5py.File(file_path)
    >>> soc = f['SOC/soc_mean']
    >>> tc = TransCom()
    >>> tc.summarize_by_transcom(np.where(soc == -9999, 0, soc), 'M09')
    {'Australia': [736.73413, 914.3717],
     'Eurasia Boreal': [3593.8977, 1461.0577],
      ...
     'Tropical Asia': [1294.8158, 1264.5378]}

Command line use:

    $ python transcom.py summarize ./files/*.h5
        --field="SOC/soc_mean"
        --output-path="./summaries.csv"

    $ python transcom.py summarize ./files/*.h5
        --field="SOC/soc_mean"
        --output-path="./summaries.csv" --summaries="('nanmean',)"

1. https://www.esrl.noaa.gov/gmd/ccgg/carbontracker/download.php
'''

import csv
import os
import re
import h5py
import numpy as np
from osgeo import gdal
from osgeo import gdalconst
from osgeo import osr
from cached_property import cached_property
from scipy.io import netcdf_file
from scipy.ndimage import zoom
from pyl4c.data.fixtures import ANCILLARY_DATA_PATHS
from pyl4c.epsg import EPSG
from pyl4c.spatial import EASE2_GRID_PARAMS, array_to_raster, as_array
from pyl4c.lib.cli import CommandLineInterface, ProgressBar
from pyl4c.lib.tcf import TYPE_MAP, TCFArray

TRANSCOM_DATA = ANCILLARY_DATA_PATHS['transcom_netcdf_path']

class TransCom(object):
    '''
    Convenience class for doing zonal statistics over TransCom regions.
    '''
    # Region codes for non-terrestrial/ off-shore regions
    offshore_region_codes = range(12, 22)
    onshore_region_codes = range(1, 12)
    onshore_region_labels = ( # i.e., regions 1 through 11, inclusive
        'North American Boreal',
        'North American Temperate',
        'South American Tropical',
        'South American Temperate',
        'Northern Africa',
        'Southern Africa',
        'Eurasia Boreal',
        'Eurasia Temperate',
        'Tropical Asia',
        'Australia',
        'Europe')
    onshore_regions = dict(zip(onshore_region_codes, onshore_region_labels))

    def __init__(self):
        f = netcdf_file(TRANSCOM_DATA, 'r')
        self.areas = f.variables['transcom_regions_area'].data.copy()
        regions = f.variables['transcom_regions'].data.copy()
        f.close()

        # TransCom array is oriented south-up, so flip it
        self.data = np.flip(regions.astype(np.int8), 0)

    def __filter__(self, arr):
        'Filters the TransCom regions array to terrestrial codes only'
        return np.where(
            ~np.isin(self.data, self.offshore_region_codes),
            self.data, 0)

    def __resample__(self, arr, shp):
        'Nearest neighbor interpolation to a new pixel size'
        # Nearest neighbor interpolation; even though ndimage's zoom()
        #   gets edge pixels wrong, it is probably the best choice here
        return resize(arr, shp, order = 0, preserve_range = True)

    @cached_property
    def __transcom_1km__(self):
        rast = self.transcom_on_ease2_grid(
            grid = 'M01', terrestrial = True)
        return self.__resample__(
            rast.ReadAsArray(), EASE2_GRID_PARAMS['M01']['shape'])

    @cached_property
    def __transcom_9km__(self):
        rast = self.transcom_on_ease2_grid(
            grid = 'M09', terrestrial = True)
        return self.__resample__(
            rast.ReadAsArray(), EASE2_GRID_PARAMS['M09']['shape'])

    @cached_property
    def __area_by_region__(self):
        'Based on a 9-km EASE-Grid (2.0), calculates area in square meters'
        # Reverse the dictionary so that labels identify region codes
        areas = dict((v, k) for k, v in self.onshore_regions.items())
        transcom = self.__transcom_9km__
        for i, label in self.onshore_regions.items():
            areas[label] = transcom[np.where(transcom == i)].shape[0] * 81e6
        return areas

    @property
    def __reported_area_by_region__(self):
        'Reports area in square meters for each TransCom region'
        # Reverse the dictionary so that labels identify region codes
        areas = dict((v, k) for k, v in self.onshore_regions.items())
        for i, label in self.onshore_regions.items():
            areas[label] = self.areas[i]
        return areas

    def as_ease2_array(self, grid = 'M01'):
        '''
        Returns a numpy.ndarray that has the specified EASE-Grid 2.0 shape,
        with the TransCom classes as values.
        '''
        if grid == 'M01':
            return self.__transcom_1km__
        elif grid == 'M09':
            return self.__transcom_9km__
        else:
            raise ValueError('Requested grid size not available')

    def as_raster(self, terrestrial = True):
        '''
        Returns the TransCom data as a `gdal.Dataset`.

        Parameters
        ----------
        terrestrial : bool
            True to filter to only terrestrial TransCom regions
            (Default: True)

        Returns
        -------
        gdal.Dataset
        '''
        # TransCom (source) spatial reference system
        wkt0 = osr.SpatialReference()
        wkt0.ImportFromEPSG(4326)
        gt0 = (-180, 1, 0, 90, 0, -1)
        return array_to_raster(
            self.__filter__(self.data) if terrestrial else self.data,
            gt0, str(wkt0))

    def count_ease2_by_transcom(
            self, array, grid = 'M01', scale = 1, text_labels = False,
            nodata = -9999):
        '''
        Calculates the number of non-NaN/ non-missing values in each TransCom
        region within the provided array.  Calls
        `TransCom.summarize_ease2_by_transcom()` as a subroutine.

        Parameters
        ----------
        array : numpy.ndarray
            Data array with the same SRS, grid size as desired for the
            TransCom data
        grid : str
            The EASE-Grid 2.0 name for which the TransCom data should be
            aligned and resampled on (e.g., "M01" or "M09")
        scale : int or float
            Optional scaling parameter to apply to the input array values,
            e.g., if the array values are (spatial) rates and should be
            scaled by the (equal) area of the grid cell
        text_labels : bool
            True to use the names of the TransCom regions instead of their
            region codes, in the output
        nodata : int or float
            NoData or Fill value in the array data to ignore

        Returns
        -------
        dict
        '''
        count = lambda arr: np.sum(np.where(np.isnan(arr), 0, 1))
        return self.summarize_ease2_by_transcom(
            array, count, grid, text_labels, nodata)

    def rescaled(self, size_degrees = 1):
        '''
        Rescales the TransCom grid (nominally 1-degree by 1-degree) to a
        scalar multiple equirectangular grid size; e.g., size_degrees = 0.5
        would enlarge the array from (180, 360) to (360, 720).

        Parameters
        ----------
        size_degrees : int or float
            The output (equirectangular) grid resolution in degrees

        Returns
        -------
        numpy.ndarray
        '''
        opts = { # Options to zoom()
            'order': 0,
            'zoom': (1/size_degrees),
            'mode': 'grid-constant',
            'grid_mode': True
        }
        return zoom(self.data, **opts)

    def summarize_ease2_by_transcom(
            self, array, summaries = dict(mean = np.nanmean, std = np.nanstd),
            grid = 'M01', scale = 1, text_labels = False, nodata = -9999):
        '''
        Calculates the mean and standard deviation of the input array values
        within each TransCom class.

        NOTE: The statistical summaries are accumulated as 32-bit floating
        point values, which is fastest, but will not be accurate for certain
        data types. If the input data array are large integers (e.g., soil
        organic carbon/ SOC values), then this should not be a problem.

        Parameters
        ----------
        array : numpy.ndarray
            Data array with the same SRS, grid size as desired for the
            TransCom data
        summaries : dict
            Dictionary of {label: function} for every summary statistic
            desired; function should be NumPy summary function, e.g.,
            nanmean, nansum, ...
        grid : str
            The EASE-Grid 2.0 name for which the TransCom data should be
            aligned and resampled on (e.g., "M01" or "M09")
        scale : int or float
            Optional scaling parameter to apply to the input array values,
            e.g., if the array values are (spatial) rates and should be
            scaled by the (equal) area of the grid cell
        text_labels : bool
            True to use the names of the TransCom regions instead of their
            region codes, in the output
        nodata : int or float
            NoData or Fill value in the array data to ignore

        Returns
        -------
        dict
            Dictionary with TransCom regions as labels and a nested Dictionary
            with a  key-value pair for each desired summary statistic.
        '''
        assert array.ndim == 2 or (array.ndim == 3 and array.shape[0] == 1), 'Can only work with 1-band raster arrays'
        if array.ndim == 3:
            array = array[0,...] # Unwrap 1-band raster arrays

        # Extract grid size in km; get the resampled TransCom array
        g = int(re.compile(r'.*(?P<km>\d{2})').match(grid).groups()[0])
        transcom = getattr(self, '__transcom_%dkm__' % g)
        assert transcom.shape == array.shape, 'Input array does not match the TransCom regions grid at the specified grid size'

        # Fill in NaN where there is NoData
        if nodata is not None:
            array = np.where(array == -9999, np.nan, array)
        # Determine how we will organize statistics by class label
        if text_labels:
            # Create, e.g., {'Australia': {}, ...}
            stats = dict([(v, dict()) for v in self.onshore_regions.values()])
        else:
            # Create, e.g., {1: {}, 2: {}, ...}
            stats = dict([(k, dict()) for k in self.onshore_regions.keys()])

        for code, label in self.onshore_regions.items():
            i = label if text_labels else code
            query = np.multiply( # Scale cell values (Default = 1.0)
                np.where(np.isin(transcom, code), array, np.nan), scale)
            for stat_name, func in summaries.items():
                # NOTE: Runs faster if dtype of accumulator is *not* set
                stats[i][stat_name] = func(query)

        return stats

    def transcom_on_ease2_grid(self, grid = 'M01', terrestrial = True):
        '''
        Projects the equirectangular TransCom regions data onto the EASE-Grid
        2.0 spatial reference system.

        NOTE: EASE-Grid 2.0 only extends to 84 degrees latitude North or
        South, so the TransCom regions ought to be restricted to this same
        extent; but it doesn't actually matter and the projection is still
        accurate. I'm making a note here in case it becomes important later:

            self.__y_coords__ = f.variables['lat'].data.copy()
            idx = np.where(np.abs(self.__y_coords__) < 84)[0]
            transcom = self.data[idx,:]

        Parameters
        ----------
        grid : str
            The EASE-Grid 2.0 name for which the TransCom data should be
            aligned and resampled on (e.g., "M01" or "M09")
        terrestrial : bool
            True to filter to only terrestrial TransCom regions

        Returns
        -------
        gdal.Dataset
        '''
        # EASE-Grid 2.0 (target) spatial reference system
        wkt = osr.SpatialReference()
        wkt.ImportFromWkt(EPSG[EASE2_GRID_PARAMS[grid]['epsg']])

        # Create a gdal.Dataset from TransCom data
        rast0 = self.as_raster(terrestrial)
        gt0 = rast0.GetGeoTransform()
        wkt0 = str(rast0.GetProjection())
        py, px = self.data.shape
        px += 12 # HACK: Output is clipped for some reason (=/)

        # The output (projected) raster's GeoTransform is difficult to
        #   determine, but this should do it automatically
        gt = gdal.AutoCreateWarpedVRT(
            rast0, str(wkt0), str(wkt), gdal.GRA_NearestNeighbour).GetGeoTransform()
        # rast0 is input raster, rast is output raster
        rast = gdal.GetDriverByName('MEM').Create('', px, py, 1, gdalconst.GDT_Int16)
        rast.SetGeoTransform(gt)
        rast.SetProjection(str(wkt))
        gdal.ReprojectImage(
            rast0, rast, str(wkt0), str(wkt), gdalconst.GRA_NearestNeighbour)
        return rast


class TransComCLI(CommandLineInterface):
    '''
    A command-line interface (CLI) for convenience; used with Google fire.
    '''
    def __init__(
            self, output_path = None, field = 'SOC/soc_mean',
            summaries = ('nanmean', 'nanstd'), grid = 'M09', **kwargs):
        self._output_path = output_path
        self._field = field
        self._grid = grid
        self._summaries = summaries
        self._kwargs = kwargs
        self._kwargs['grid'] = grid
        self.tc = TransCom()

    def __check__(self):
        assert self._output_path is not None,\
            'You must specify an output_path with: --output-path=""'
        assert os.path.exists(os.path.dirname(self._output_path)),\
            'Did not recognize output_path (Cannot use shortcuts like ~)'

    def __dump_csv__(self, output_path, items):
        with open(output_path, 'w') as stream:
            writer = csv.writer(stream, delimiter = ',', quotechar = '"')
            writer.writerow(('filename', 'transcom_region', 'statistic', 'value'))
            for each in items:
                writer.writerow(each)

    def __expand_summary__(self, items):
        'Expands ["A", {"a": 1, "b": 2}] into [("A", "a", 1), ("A", "b", 2)]'
        new_items = []
        # TODO Would be nice to generalize this using a recursive pattern
        for filename, d in items:
            for label, summary in d.items():
                for key, value in summary.items():
                    new_items.append((filename, label, key, value))
        return new_items

    def report_areas(self):
        '''
        Prints the reported area of each TransCom region.
        '''
        # NOTE: The underlying data are 32-bit floating point, so this
        #   string formatting will maximize the precision
        for k, v, in self.tc.__reported_area_by_region__.items():
            print('%s: %.0f' % (k, v))

    def summarize(self, *file_paths):
        '''
        Creates a statistical summary by class label (TransCom region) for
        each of multiple input HDF5 files. Use:

            $ python transcom.py summarize ./files/*.h5 --field="SOC/soc_mean"
                --output-path="./summaries.csv"

            $ python transcom.py summarize ./files/*.h5 --field="SOC/soc_mean"
                --output-path="./summaries.csv" --summaries="('nanmean',)"

        Parameters
        ----------
        *file_paths : str
        field : str
            Hierarchical path name of field to summarize: `--field="<field>"`
        summaries : str
            For compatibility at the CLI, can describe the statistical summary
            functions desired as a comma-delimited string, e.g.,
            `"('nanmean','nanstd')"` where names refer to NumPy functions
        output_path : str
            File path where the output CSV file should be written
        **kwargs : str
            Other keyword arguments to be passed on to
            `TransCom.summarize_ease2_by_transcom()`
        '''
        self.__check__()
        results = []
        n = len(file_paths)
        assert n > 0, 'No file paths given (Did you forget to specify "field" argument?)'
        # NOTE: fire.Fire() implicitly transforms comma-delimited string into
        #   tuple; here, e.g., "('nanmean', 'nanstd')" becomes:
        #   {'nanmean': np.nanmean, 'nanstd': np.nanstd}
        self._kwargs['summaries'] = dict([
            (name, getattr(np, name)) for name in self._summaries
        ])

        # Determine what kind of files we're working with
        mode = 'other'
        if file_paths[0].split('.')[-1] == 'h5':
            mode = 'hdf5'
        elif file_paths[0].split('.')[-1] in TYPE_MAP.keys():
            mode = 'sparse'

        with ProgressBar(len(file_paths), 'Summarizing files...') as progress:
            for i, filename in enumerate(file_paths):
                if mode == 'hdf5':
                    with h5py.File(filename, 'r') as hdf:
                        arr = hdf[self._field][:]

                elif mode == 'sparse':
                    tcf = TCFArray(filename, self._grid)
                    tcf.inflate()
                    arr = tcf.data

                elif mode == 'other':
                    arr, _, _ = as_array(filename, band_axis = False)

                stats = self.tc.summarize_ease2_by_transcom(arr, **self._kwargs)
                results.append((filename, stats))
                progress.update(i + 1)

        self.__dump_csv__(self._output_path, self.__expand_summary__(results))


if __name__ == '__main__':
    # Performance notes:
    #   - Does run faster if the dtype of the accumulator function
    #       (e.g., np.nanmean) is NOT set
    #   - Does NOT run faster if arrays are raveled prior to summary
    import fire
    fire.Fire(TransComCLI)



'''
Applications of the SMAP Level 4 Carbon (L4C) model, including sub-modules
related to L4C model calibration (`pyl4c.apps.calibration`) and forward run
simulation (`pyl4c.apps.l4c`). **Notably, down-scaling of L4C data from 9-km
resolution to 1-km resolution can be performed in with
`pyl4c.apps.resample`.**
'''



'''
SMAP Level 4 Carbon (L4C) forward run logic, including data structures for
running an L4C forward simulation. A point (site-level) version of the L4C
algorithm can be examined and run in `pyl4c.apps.l4c.main`.
'''

import csv
import datetime
import warnings
import numpy as np
from functools import partial
from pyl4c import suppress_warnings
from pyl4c.lib.cli import ProgressBar

class AbstractData(object):
    '''
    An abstraction of a data container for running L4C simulations.
    Represents discontinuous locations or "points" on a global EASE-Grid 2.0
    over which to run an L4C simulation. Each "point" has a 1-km subgrid
    associated with the 9-km "point" cell.
    '''
    @property
    def labels(self):
        return self._data_labels

    @labels.setter
    def labels(self, new_labels):
        self._data_labels = tuple(new_labels)

    @property
    def shape(self):
        if self.data is not None:
            return self.data.shape


class L4CConstants(object):
    '''
    A namespace for constant values used in running L4C. These constant
    values are broadcast to an array that is conformable with driver data
    for fast and seamless computation in the forward run.

    Parameters
    ----------
    count : int
        Number of point locations ("sites") used in forward simulation
    pft_array : numpy.ndarray
        Array specifying the PFT at each point location
    valid_pft : Sequence
        The valid PFT codes
    '''
    def __init__(self, count : int, pft_array, valid_pft = range(1, 10)):
        # Set invalid PFT codes to 0
        self._count = count
        self._pft = np.where(np.in1d(pft_array.ravel(), valid_pft),
            pft_array.ravel(), 0)

    def conformable(self, array):
        '''
        Creates a `(1) x N x P` array where `P` is the number of PFT classes.
        Useful for guaranteeing that an array is conformable with the other
        arrays it will be concatenated with.

        Parameters
        ----------
        array : numpy.ndarray
        '''
        assert array.ndim <= 4, 'Expected ndim <= 4'
        if array.ndim == 1:
            # A 1D array indicates one value per PFT
            return np.apply_along_axis(
                lambda p: array[p], 0, self._pft
            ).reshape((self._count, 81))
        if array.ndim == 2:
            if array.shape == (self._count, 81):
                return array # Array already conforms
            # A 2D array indicates multiple ("ndim") values per PFT
            return np.apply_along_axis(
                lambda p: array[:,p], 0, self._pft
            ).reshape((array.shape[0], self._count, 81))
        if array.ndim == 3:
            # i.e., a 365-day climatology
            assert array.shape == (365, self._count, 81),\
                'No support for 3D arrays that do not have 365 elements on axis 0 (i.e., a 365-day climatology)'
            return array
        if array.ndim == 4:
            assert array.shape[-3:] == (365, self._count, 81),\
                'No support for 4D arrays that do not have the following sub-space: (%s)' % ' x '.join(map(str, array.shape[-3:]))
            return array

    def add(self, array, label):
        '''
        Adds a new constant as an attribute on this instance.

        Parameters
        ----------
        array : numpy.ndarray
            The array to add
        label : str
            The name by which this array will be known
        '''
        setattr(self, label, self.conformable(array))


class L4CDrivers(AbstractData):
    '''
    Represents a collection of L4C driver variables. Driver variables *drive*
    the model but are not in themselves updated; they are exogenous variables
    with values fixed at the start of the simulation.

    Parameters
    ----------
    count : int
        Number of point locations ("sites") used in forward simulation
    drivers
    labels
    '''
    def __init__(self, count, drivers = None, labels = None):
        if drivers is not None and labels is not None:
            if hasattr(drivers, 'ndim'):
                matched = drivers.shape[0] == len(labels)
            else:
                matched = len(drivers) == len(labels)
            if not matched:
                raise ValueError('There should be as many labels as there are driver variables')
        self._count = count
        self._data = None
        self._data_labels = labels
        # Use appropriate setter function
        self.data = drivers

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, new):
        if new is None:
            self._data = None
        elif hasattr(new, 'ndim'):
            self._data = new
        else:
            # Otherwise, concat the sequence of arrays, once made conformable
            self._data = np.concatenate(list(map(
                partial(self.conformable), new)), axis = 0)

    def conformable(self, array):
        '''
        Given an array, returns representation that conforms to expectations.

        Parameters
        ----------
        array : numpy.ndarray
            Input data array

        Returns
        -------
        numpy.ndarray
            `(1 x T x N x 81)` array
        '''
        shp = array.shape
        assert (self._count in shp),\
            'Array is not conformable with %d measurement points' % self._count
        assert array.ndim in (2, 3, 4),\
            'Driver variables should have 2, 3, or 4 axes only'

        # Driver data should be (1 x T x N x 81)
        i_idx = shp.index(self._count)
        if array.ndim == 2:
            t = shp[0] if i_idx == 1 else shp[1] # Get size of time axis
            return array.swapaxes(1, i_idx).reshape((t, self._count, 1))\
                    .repeat(81, axis = 2).reshape((1, t, self._count, 81))

        # NOTE: The swapaxes() calls essentially assert place each axis into
        #   its desired position, moved from its inferred position
        j_idx = shp.index(81) if 81 in shp else shp.index(1) # Subgrid axis
        t_idx = set((0, 1, 2)).difference((i_idx, j_idx)).pop() # Time axis
        if array.ndim == 3:
            t = shp[t_idx] # Get the size of the time axis
            if shp[j_idx] == 81:
                return array.swapaxes(1, i_idx).swapaxes(2, j_idx)\
                    .reshape((1, t, self._count, 81))
            return array.swapaxes(1, i_idx).swapaxes(2, j_idx)\
                .reshape((1, t, self._count, 1)).repeat(81, axis = 3)

        if array.ndim == 4:
            return array.swapaxes(1, t_idx).swapaxes(2, i_idx)\
                .swapaxes(3, j_idx)


class L4CState(AbstractData):
    '''
    Represents a collection of L4C state variables. State variables influence
    the driving of the model; an initial state (at time t=0) is required to
    start running and the variable state might be updated at each time step
    (t) based on the value at step (t-1). State data are stored as
    `M x T x N x 81` arrays.

    Parameters
    ----------
    count : int
        Number of point locations ("sites") used in forward simulation
    state : numpy.ndarray or None
        (Optional) Initial state
    labels : list or tuple or None
        (Optional) Sequence of labels for each state variable, should be
        the same length as the first axis of the `state` of array
    axis_labels : list or tuple or None
        Names of each axis of the stored state matrix; axes are, in order:
        (state variable, time, site, 81)
    dtype
        The data type to coerce on the state array
    '''
    def __init__(
            self, count : int, state = None, labels = None,
            axis_labels = None, dtype = np.float32):
        if state is not None and labels is not None:
            if hasattr(state, 'ndim'):
                matched = state.shape[0] == len(labels)
            else:
                matched = len(state) == len(labels)
            if not matched:
                raise ValueError('There should be as many labels as there are state variables')
        self._axis_labels = axis_labels
        self._count = int(count)
        self._bounds = (0, np.inf)
        self._data = None
        self._data_labels = labels
        self._dtype = dtype
        # Use appropriate setter function
        if state is not None:
            self.data = state.astype(dtype)

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, new):
        if new is None:
            self._data = None
        elif hasattr(new, 'ndim'):
            assert new.ndim == 4,\
                'New data array must have 4 axes; should be (M x T x N x 81)'
            self._data = new.astype(self._dtype)
        else:
            # Otherwise, concat the sequence of arrays, once made conformable
            self._data = np.concatenate(
                list(map(self.conformable, new)), axis = 0).astype(self._dtype)

    def _serialize(
            self, state, label, output_path, truncate, time_labels, prec,
            verbose = False):
        'Argument "state" should be length-T sequence of length-N sequences'
        now = datetime.datetime.now().strftime('%Y-%m-%d')
        site_labels = list(range(0, self._count))
        if self._axis_labels is not None:
            site_labels = self._axis_labels[-2]
            # Decode UTF-8 strings
            if hasattr(site_labels[0], 'decode'):
                site_labels = list(map(lambda x: x.decode(), site_labels))
        with open(output_path, 'w') as file:
            writer = csv.writer(file)
            writer.writerow(
                ('# L4C Model State %s serialized %s' % (label, now),))
            writer.writerow(('site', 'time', 'value'))
            with ProgressBar(len(state), verbose = verbose) as progress:
                if verbose:
                    progress.prefix = 'Serializing...'
                for t in range(0, len(state)):
                    for i, label in enumerate(site_labels):
                        # Working with a length-T sequence of length-N lists
                        writer.writerow([
                            label, time_labels[t],
                            format(state[t,i], '.%df' % prec)
                        ])
                    if verbose:
                        progress.update(t)

    @suppress_warnings
    def _serialize_combined(
            self, label, output_path, truncate, time_labels, prec):
        'Serializes the combined (summed) state, summing across fields'
        # Sum across fields, then take the mean across the 1-km subgrid
        state = np.nanmean(self.data.sum(axis = 0), axis = -1)
        if truncate is not None:
            state = state[0:truncate,...].tolist()
        else:
            state = state.tolist()
        self._serialize(
            state, label, output_path, truncate, time_labels, prec)

    def advance(self, label, t, delta = None, func = np.add, bounds = None):
        '''
        Updates the state data at time (t) based on the prior state (state
        at t-1). NOTE: "delta" is the left argument of the binary function.

        Parameters
        ----------
        label : str
            One of the labels in self.labels
        t : int
            Index of the time step to update
        delta : numpy.ndarray
            LHS of the function (func) to change state (Default: None)
        func : function
            The binary function to use to combine delta and the state
            at time (t-1) (Default: numpy.add).
        bounds : dict
        '''
        i = self.labels.index(label)
        if delta is not None:
            new_state = func(delta, self._data[i, t-1, ...])
            # Check for out-of-bounds
            if bounds is None:
                self._data[i, t, ...] = new_state
                return

            self._data[i, t, ...] = np.where(
                new_state > bounds[1], bounds[1], np.where(
                    new_state < bounds[0], bounds[0], new_state))

    def conformable(self, array, allocation = 1):
        '''
        Given an array, returns representation that conforms to expectations.

        Parameters
        ----------
        array : numpy.ndarray
            Input data array
        allocation : int
            Number of time steps to allocate

        Returns
        -------
        numpy.ndarray
            `(1 x T x N x 81)` array
        '''
        shp = array.shape
        assert (self._count in shp),\
            'Array is not conformable with %d measurement points' % self._count
        assert array.ndim in (2, 3),\
            'State variables should have either 2 or 3 axes only'

        # We want to return a (1 x T x N x 81) array; tests are somewhat
        #   verbose in light of the possibility that self._count = 81
        assert 81 in shp,\
            'State variables must have a 1-km subgrid axis with 81 elements'
        i_idx = shp.index(self._count)
        j_idx = shp.index(81)
        # Add concatenation (axis=0) and time (axis=1) axes, then allocate
        #   T time steps
        if array.ndim == 2:
            return np.concatenate((
                array.swapaxes(i_idx, 0).reshape((1, 1, self._count, 81)),
                np.full((1, allocation - 1, self._count, 81), np.nan)
            ), axis = 1)
        if array.ndim == 3:
            return np.concatenate((
                array.swapaxes(i_idx, 1).swapaxes(j_idx, 2)\
                    .reshape((1, 1, self._count, 81)),
                np.full((1, allocation - 1, self._count, 81), np.nan)
            ), axis = 1)

    def update(self, label, t, array, bounds = None):
        '''
        Updates the state data with the given label by inserting an array
        at the specified time step (t).

        Parameters
        ----------
        label : str
            One of the labels in self.labels
        t : int
            Index of the time step to update
        array : numpy.ndarray
        bounds : dict
        '''
        i = self.labels.index(label)
        if bounds is None:
            self._data[i, t, ...] = array
            return

        self._data[i, t, ...] = np.where(
            array > bounds[1], bounds[1], np.where(
                array < bounds[0], bounds[0], array))

    def serialize(
            self, output_tpl, truncate = None, time_labels = None, prec = 3,
            verbose = False):
        '''
        Dumps the current state to a CSV file. Data on the 1-km subgrid are
        summarized; the mean at each 9-km site is reported.

        Parameters
        ----------
        output_tpl : str
            Template filename for the output CSV file
        truncate : int
            (Optional) Largest time index that will be dumped
        time_labels : list or tuple
            (Optional) A sequence of time labels
        prec : int
            Number of decimal places to include (Default: 1)
        verbose : bool
            True to print progress of serialization
        '''
        assert truncate is None or (hasattr(truncate, 'real') and not hasattr(truncate, 'is_integer')), '"truncate" must be an integer'
        # Insert a string formatting character for the state's label
        output_tpl = '.'.join(output_tpl.split('.')[:-1])
        output_tpl = '%s_%%s.csv' % output_tpl

        if time_labels is None:
            time_labels = range(0, self.shape[1])
        else:
            assert isinstance(time_labels[0], str),\
                'time_labels must be str type'

        # Check that there are at least as many time labels as time indices
        t_max = truncate if truncate is not None else self.shape[1]
        assert len(time_labels) >= t_max, 'Not enough time_labels!'

        # If state represents SOC, combine across the SOC pools
        if all([s.startswith('soc') for s in self.labels]):
            self._serialize_combined(
                'SOC', output_tpl % 'SOC', truncate, time_labels, prec)
            return

        # Take the mean across the 1-km subgrid
        state = self.data
        if self.data.ndim > 3:
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                state = np.nanmean(self.data, axis = -1)
        print('Datasets to serialize: %d' % self.shape[0])
        for i in range(0, self.shape[0]):
            self._serialize(
                state[i,...], self.labels[i].upper(),
                output_tpl % self.labels[i].upper(), truncate, time_labels,
                prec, verbose = verbose)


class L4CStratifiedState(L4CState):
    '''
    Represents a collection of L4C state variables that are potentially
    stratified into vertical layers. State data are stored as
    `M x T x Z x N x 81` arrays, where `M` enumerates the different state
    variables, `Z` enumerates the depths, `T` is the time axis, and `N`
    enumerates the number of sites.

    Parameters
    ----------
    layers : int
        Number of vertical layers
    count : int
        Number of point locations ("sites") used in forward simulation
    state : numpy.ndarray or None
        (Optional) Initial state
    labels : list or tuple or None
        (Optional) Sequence of labels for each state variable, should be
        the same length as the first axis of the `state` of array
    axis_labels : list or tuple or None
        Names of each axis of the stored state matrix; axes are, in order:
        (state variable, time, site, 81)
    dtype
        The data type to coerce on the state array
    '''
    def __init__(
            self, layers: int, count: int, state = None, labels = None,
            axis_labels = None, dtype = np.float32):
        self._layers = int(layers)
        if axis_labels is not None:
            assert len(axis_labels) == 5,\
                '"axis_labels" should contain one sequence for each axis (5 axes)'
        super().__init__(
            count = count, state = state, labels = labels,
            axis_labels = axis_labels, dtype = dtype)

    def _serialize(
            self, state, label, output_path, truncate, time_labels, prec,
            verbose = False):
        'Argument "state" should be length-T sequence of length-N sequences'
        now = datetime.datetime.now().strftime('%Y-%m-%d')
        site_labels = list(range(0, self._count))
        if self._axis_labels is not None:
            _, _, depth_labels, site_labels, _ = self._axis_labels
            # Decode UTF-8 strings
            if hasattr(site_labels[0], 'decode'):
                site_labels = list(map(lambda x: x.decode(), site_labels))
        with open(output_path, 'w') as file:
            writer = csv.writer(file)
            writer.writerow(
                ('# L4C Model State %s serialized %s' % (label, now),))
            writer.writerow(('site', 'depth', 'time', 'value'))
            with ProgressBar(len(state), verbose = verbose) as progress:
                if verbose:
                    progress.prefix = 'Serializing...'
                for t in range(0, len(state)):
                    for z, depth in enumerate(depth_labels):
                        for i, site_name in enumerate(site_labels):
                            # Working with a length-T sequence of length-N lists
                            writer.writerow([
                                site_name, depth, time_labels[t],
                                format(state[t][z][i], '.%df' % prec)
                            ])
                        if verbose:
                            progress.update(t)

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, new):
        if new is None:
            self._data = None
        elif hasattr(new, 'ndim'):
            assert new.ndim >= 4,\
                'New data array must have at least 4 axes; should be (T x Z x N x 81)'
            self._data = new.astype(self._dtype)
        else:
            # Otherwise, concat the sequence of arrays, once made conformable
            self._data = np.concatenate(
                list(map(self.conformable, new)), axis = 0).astype(self._dtype)

    def conformable(self, array, allocation = 1):
        '''
        Given an array, returns representation that conforms to expectations.

        Parameters
        ----------
        array : numpy.ndarray
            Input data array
        allocation : int
            Number of time steps to allocate

        Returns
        -------
        numpy.ndarray
            `(1 x T x Z x N x 81)` array
        '''
        shp = array.shape
        assert (self._count in shp),\
            'Array is not conformable with %d measurement points' % self._count
        assert array.ndim in (3, 4, 5),\
            'State variables should have between 3 and 5 axes only'

        # We want to return a (1 x Z x T x N x 81) array; tests are somewhat
        #   verbose in light of the possibility that self._count = 81
        assert 81 in shp,\
            'State variables must have a 1-km subgrid axis with 81 elements'
        i_idx = shp.index(self._count)
        j_idx = shp.index(81)
        if array.ndim == 3:
            return np.concatenate((
                array.swapaxes(1, i_idx).swapaxes(2, j_idx)\
                    .reshape((1, 1, self._count, 81)),
                np.full((1, allocation - 1, 1, self._count, 81), np.nan)
            ), axis = 1)
        # Arrays of 4D or higher are assumed to have a Z axis
        z_idx = shp.index(self._layers) if self._layers in shp else shp.index(1)
        if array.ndim == 4:
            return np.concatenate((
                array.swapaxes(0, z_idx).swapaxes(2, i_idx).swapaxes(3, j_idx)\
                    .reshape((1, self._layers, 1, self._count, 81)),
                np.full((1, allocation - 1, self._layers, self._count, 81), np.nan),
            ), axis = 1)
        if array.ndim == 5:
            return np.concatenate((
                array.swapaxes(0, z_idx).swapaxes(3, i_idx).swapaxes(4, j_idx)\
                    .reshape((1, self._layers, 1, self._count, 81)),
                np.full((1, allocation - 1, self._layers, self._count, 81), np.nan),
            ), axis = 1)


def report(hdf):
    '''
    Check that we have everything needed to run L4C, print a report to the
    screen.

    Parameters
    ----------
    hdf : h5py.File
    '''
    KEYS = ('apar', 'vpd', 'ft', 'tmin', 'tsoil', 'smrz', 'smsf')

    def find(hdf, prefix, key, pad = 10):
        'Find a key, print the report'
        try:
            field = '%s/%s' % (prefix, key)
            if len(hdf[field].shape) == 2 or key == 'fpar':
                pretty = ('"%s"' % key).ljust(pad)
                print_stats(hdf[field][:], pad, pretty)
            elif len(hdf[field].shape) == 3:
                # Assuming data are enumerated on the first axis
                for i in range(0, hdf[field].shape[0]):
                    pretty = ('"%s" (%d)' % (key, i)).ljust(pad)
                    print_stats(hdf[field][i,...], pad, pretty)
        except KeyError:
            pretty = ('"%s"' % key).ljust(pad)
            print('-- MISSING %s' % pretty)

    def print_stats(data, pad, pretty):
        shp = ' x '.join(map(str, data.shape))
        shp = ('[%s]' % shp).ljust(pad + 7)
        stats = tuple(summarize(data))
        stats_pretty = ''
        if stats[0] is not None:
            stats_pretty = '[%.2f, %.2f]' % (stats[0], stats[2])
            if len(key) < 10:
                print('-- Found %s %s %s' % (pretty, shp, stats_pretty))
            else:
                print('-- Found %s' % pretty)
                print('%s%s %s' % (''.rjust(pad + 10), shp, stats_pretty))

    def summarize(data, nodata = -9999):
        'Get summary statistics for a field'
        if str(data.dtype).startswith('int'):
            return (None for i in range(0, 3))
        data[data == -9999] = np.nan
        return (
            getattr(np, f)(data) for f in ('nanmin', 'nanmean', 'nanmax')
        )

    print('\nL4C: Validating configuration and input datasets for file:')
    print('  %s' % hdf.filename)
    print('\nL4C: Checking for required driver variables...')
    for key in KEYS:
        if key == 'ft' and key not in hdf['drivers'].keys():
            find(hdf, 'drivers', 'tsurf')
        elif key == 'apar' and key not in hdf['drivers'].keys():
            find(hdf, 'drivers', 'par')
            find(hdf, 'drivers', 'fpar')
        else:
            find(hdf, 'drivers', key)

    print('\nL4C: Checking for required state variables...')
    for key in ('PFT', 'npp_sum', 'soil_organic_carbon',):
        find(hdf, 'state', key)

    print('\nL4C: Summarizing metadata...')
    y1, m1, d1, _ = hdf['time'][0,...]
    y2, m2, d2, _ = hdf['time'][-1,...]
    print('-- First date: %s' % datetime.datetime(y1, m1, d1)\
        .strftime('%Y-%m-%d'))
    print('-- Final date: %s' % datetime.datetime(y2, m2, d2)\
        .strftime('%Y-%m-%d'))
    print('-- Total length: %d' % hdf['time'].shape[0])
    print('')



'''
Tools for working with L4SM Nature Run data.

TODO: Refactor to use `numpy.ravel_multi_index`?
'''

import numpy as np
import netCDF4
from cached_property import cached_property
from pyl4c.data.fixtures import EASE2_GRID_PARAMS

class NatureRunNetCDF4(object):
    '''
    Represents an L4SM Nature Run netCDF4 dataset. Contains methods for
    converting from the compact, 1D "tile-space" to a 2D EASE-Grid 2.0.

    Would have liked to subclass netCDF4.Dataset directly, but it is part of
    a Cython library so I can't interrogate its class structure.

    Parameters
    ----------
    file_path : str
        The NetCDF4 file path
    '''
    def __init__(self, file_path: str):
        self._dataset_file_path = file_path
        self._grid = 'M09'
        try:
            self.dataset = netCDF4.Dataset(file_path)
        except OSError:
            raise OSError('Due to a bug in netCDF4, you MUST import netCDF4 BEFORE h5py; import this module (calibration) first if using in a script')

    @property
    def variables(self):
        return self.dataset.variables

    @cached_property
    def _col_idx(self):
        return np.array(self.dataset.variables['IG'][:])

    @cached_property
    def _row_idx(self):
        return np.array(self.dataset.variables['JG'][:])

    @classmethod
    def index_bulk(cls, nc, variable, tile_idx, reducer = 'mean'):
        '''
        A more efficient way of indexing, assuming that tile indices are
        already known. This avoids the expensive mapping of the same set
        of EASE-Grid 2.0 coordinates to tile space, and can be applied
        to multiple tile-space netCDF4 files.

        Parameters
        ----------
        nc : netCDF4.Dataset
            The netCDF4 file to sample from
        variable : str
            Name of the variable
        tile_idx : list or tuple
            Tile-space indices

        Returns
        -------
        numpy.ndarray
        '''
        target_variable = nc.variables[variable]
        # Minimize the extent of the array we need to read from the file;
        #   e.g., tile space indices are [0, 1, ... N] but we may only need
        #   to index between A and B, so instead of reading from 0 all the way
        #   to N, just read between A (tmin) and B (tmax).
        tmin, tmax = (min(tile_idx), max(tile_idx) + 1)
        subarray = target_variable[:,tmin:tmax]
        # Check if we're working with a 3-hourly L4SM variable
        if target_variable.ndim == 2:
            value_store = [
                getattr(subarray[:,idx - tmin], reducer)(axis = 0)
                for idx in tile_idx
            ]
        else:
            value_store = [
                subarray[idx - tmin] for idx in tile_idx
            ]
        return np.array(value_store)

    def as_tile(self, array):
        'Converts a 2D EASE-Grid 2.0 array to tile format'
        new_array = np.ones(self._row_idx.shape) * np.nan
        for i in range(0, self._row_idx.shape[0]):
            row, col = self.tile_to_ease2(i)
            new_array[i] = array[row, col]
        return new_array

    def tile_to_ease2(self, tile_idx):
        'Maps 1D tile coordinates to 2D EASE-Grid 2.0 coordinates'
        return (self._row_idx[tile_idx], self._col_idx[tile_idx])

    def ease2_to_tile(self, row, col):
        'Maps 2D EASE-Grid 2.0 coordinates to 1D tile coordinates'
        length = self._row_idx.size
        match = np.logical_and(self._row_idx == row, self._col_idx == col)
        if not np.any(match):
            raise ValueError("No matching row-column index pair was found; requested cell might be outside of the model domain")
        return int(np.arange(length)[match])

    def index(self, variable, indices):
        '''
        Returns values at the given indices, based on row-column pairs
        from a 2D EASE-Grid 2.0.

        NOTE: It is MUCH faster (by an order of magnitude) to use the class
        method index_bulk() if you need to extract the same indices from
        multiple tile-space netCDF4 files. The recommended pattern is based
        on this function's signature, where the `tile_idx` are computed
        first and then passed to index_bulk().

        Parameters
        ----------
        variable : str
            Name of the netCDF4 variable to inflate
        indices : tuple or list
            Sequence of (row, column) 2-tuples

        Returns
        -------
        numpy.ndarray
        '''
        assert hasattr(indices, 'index'), 'Indices should be a sequence of (row, column) pairs'
        target_variable = self.dataset.variables[variable]
        tile_idx = [ # This is what takes the longest
            self.ease2_to_tile(row, col) if row is not None and col is not None else None
            for row, col in indices
        ]
        return self.__class__.index_bulk(self, variable, tile_idx)

    def inflated(self, variable, nodata = -9999):
        '''
        Returns a "tile space" variable (1D) as a 2D NumPy array, calculating
        a daily average as needed.

        Parameters
        ----------
        variable : str
            Name of the netCDF4 variable to inflate
        nodata : int or float
            The NoData value to use in the output array

        Returns
        -------
        numpy.ndarray
        '''
        target_variable = self.dataset.variables[variable]
        shp = EASE2_GRID_PARAMS[self._grid]['shape']
        result = np.ones(shp) * nodata

        # Check if we're working with a 3-hourly L4SM variable
        if target_variable.ndim == 2 and target_variable.shape[0] == 8:
            avg = np.array(target_variable[:]).mean(axis = 0)
            for i in range(0, target_variable.shape[1]):
                # Calculate daily mean (collapse first axis)
                row, col = self.tile_to_ease2(i)
                result[row, col] = avg[i]
        else:
            for i in range(0, target_variable.shape[0]):
                row, col = self.tile_to_ease2(i)
                result[row, col] = target_variable[i]

        return result



'''
This is a collection of Python tools for managing, analyzing, and visualizing
data from the NASA Soil Moisture Active Passive (SMAP) Level 4 Carbon (L4C)
mission. In particular:

- Working with data in EASE-Grid 2.0 projection (`pyl4c.ease2`;)
- Converting HDF5 geophysical variables to GeoTIFF format (`pyl4c.spatial`);
- Creating statistical summaries of SMAP L4C variables or other raster arrays (`pyl4c.utils`);
- Reproducing L4C operational model logic (`pyl4c.science`);
- Down-scaling L4C flux and SOC state variables (`pyl4c.apps.resample`);
- Calibrating the L4C model (`pyl4c.apps.calibration`);
- Running the L4C model (`pyl4c.apps.l4c`);
- Aligning and summarizing SMAP L4C variables with TransCom regions (`pyl4c.lib.transcom`);

**Key things to note:**

- File paths for your system may need to be updated in order to access
    essential ancillary datasets; see `pyl4c.data.fixtures`.
'''

__pdoc__ = {}
__pdoc__['tests'] = False

from collections import Counter
import warnings
import numpy as np

class Namespace(object):
    '''
    Dummy class for holding attributes.
    '''
    def __init__(self):
        pass

    def add(self, label, value):
        '''
        Adds a new attribute to the Namespace instance.

        Parameters
        ----------
        label : str
            The name of the attribute
        value : None
            Any kind of value to be stored
        '''
        setattr(self, label, value)


def equal_or_nan(x, y):
    '''
    A variation on numpy.equal() that also returns True where respective
    inputs are NaN.

    Parameters
    ----------
    x : numpy.ndarray
    y : numpy.ndarray

    Returns
    -------
    numpy.ndarray
    '''
    return np.where(
        np.logical_or(np.isnan(x), np.isnan(y)), True, np.equal(x, y))


def haversine(p1, p2, radius = 6371e3):
    '''
    Haversine formula for great circle distance, in meters. Accurate for
    points separated near and far but for small distances the accuracy is
    improved by providing a different radius of the sphere, say 6356.7523 km
    for polar regions or 6378.1370 km for equatorial regions. Default is the
    mean earth radius.

    NOTE: Distance returned is in the same units as radius.

    Parameters
    ----------
    p1 : tuple or list
        Sequence of two floats, longitude and latitude, respectively
    p2 : tuple or list
        Same as p1 but for the second point
    radius : int or float
        Radius of the sphere to use in distance calculation
        (Default: 6,371,000 meters)

    Returns
    -------
    float
    '''
    x1, y1 = map(np.deg2rad, p1)
    x2, y2 = map(np.deg2rad, p2)
    dphi = np.abs(y2 - y1) # Difference in latitude
    dlambda = np.abs(x2 - x1) # Difference in longitude
    angle = 2 * np.arcsin(np.sqrt(np.add(
        np.power(np.sin(dphi / 2), 2),
        np.cos(y1) * np.cos(y2) * np.power(np.sin(dlambda / 2), 2)
    )))
    return float(angle * radius)


def pft_dominant(pft_map_array):
    '''
    Returns the PFT class dominant among a (1-km) subarray; for example,
    for each 9-km EASE-Grid 2.0 pixel, returns the PFT class that is dominant
    among the 1-km sub-array pixels.

    Parameters
    ----------
    pft_map_array : numpy.ndarray
        An M x N array specifying the PFT code for every pixel

    Returns
    -------
    list
        An M-element list of the dominant PFT among N pixels
    '''
    return [ # Skip invalid PFT codes
        list(filter(lambda x: x[0] in range(1, 9), count))[0][0]
        for count in [ # Count 1-km cells by PFT
            a.most_common() for a in np.apply_along_axis(
                lambda x: Counter(x.tolist()), 1, pft_map_array)
        ]
    ]


def pft_selector(pft_map, pft):
    '''
    For a given PFT class, returns the tower sites, as rank indices, that
    represent that PFT. Exceptions are made according to the L4C calibration
    protocol, e.g., sites with any amount of Deciduous Needleleaf (DNF) in
    their 1-km subgrid are considered to represent the DNF PFT class.

    Parameters
    ----------
    pft_map : numpy.ndarray
        An (N x M) array where N is the number of sites and M is the number
        of replicates within each site; e.g., for SMAP L4C, M=81 corresponding
        with the 81 cells of the 1-km subgrid for each eddy covariance flux
        tower site.
    pft : int
        The integer number of the PFT to select

    Returns
    -------
    numpy.ndarray
    '''
    idx = pft_map.shape[0]
    if pft == 3:
        return np.apply_along_axis(lambda x: x == 3, 1, pft_map).any(axis = 1)
    return np.equal(pft, [ # Skip invalid PFT codes
        count[1][0] if count[0][0] not in range(1, 9) else count[0][0]
        for count in [ # Count 1-km cells by PFT
            a.most_common() for a in np.apply_along_axis(
                lambda x: Counter(x.tolist()), 1, pft_map)
        ]
    ])


def suppress_warnings(func):
    'Decorator to suppress NumPy warnings'
    def inner(*args, **kwargs):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            return func(*args, **kwargs)
    return inner



'''
Functions related to calculating (evapo)transpiration. These are largely taken
from the MODIS MOD16 framework, but the goal is not to reproduce MOD16;
rather, they facilitate the calculation of transpiration as one of the three
components of evapotranspiration. Various primary sources are cited in the
MOD16 code:

- Monteith, J. L., and M. Unsworth. 2001. "Principles of Environmental
    Physics", Second Ed.
- Mu, Q., Zhao, M., & Running, S. W. (2011). Improvements to a MODIS
  global terrestrial evapotranspiration algorithm. *Remote Sensing of
  Environment*, 115(8), 1781–1800.

**NOTE:** Currently, canopy evaporation and potential transpiration are
denominated in seconds (i.e., kg m-2 sec-1) instead of in days, as in MOD16.
This is for compatibility with soil water infiltration models that run in
with sub-daily time steps. As such, net radiation to the land surface or
canopy should be in units of J m-2 s-1.
'''

import numpy as np

STEFAN_BOLTZMANN = 5.67e-8 # Stefan-Boltzmann constant, W m-2 K-4
SPECIFIC_HEAT_CAPACITY_AIR = 1013 # J kg-1 K-1, Monteith & Unsworth (2001)
# Ratio of molecular weight of water vapor to that of dry air (ibid.)
MOL_WEIGHT_WET_DRY_RATIO_AIR = 0.622

# Calculate the latent heat of vaporization (J kg-1)
latent_heat_vapor = lambda temp_k: (2.501 - 0.002361 * (temp_k - 273.15)) * 1e6

def canopy_evaporation(
        pressure, temp_k, rhumidity, vpd, lai, fpar, rad_canopy, f_wet = None,
        g_h = 0.01, g_e = 0.01):
    r'''
    Wet canopy evaporation calculated according to the MODIS MOD16 framework,
    with air density calculation from the [
    National Physical Laboratory (2021),
    "Buoyancy Correction and Air Density Measurement."
    ](http://resource.npl.co.uk/docs/science_technology/mass_force_pressure/
        clubs_groups/instmc_weighing_panel/buoycornote.pdf)

    $$
    \lambda E = \frac{(s\, A_c\, F_c + \rho\, C_p\, D\, (F_c / r_a))F_{wet}}{s +
        (P_a\, C_p\, r_c)(\lambda\, \varepsilon\, r_a)^{-1}}
    $$

    Parameters
    ----------
    pressure : float or numpy.ndarray
        The air pressure in Pascals
    temp_k : float or numpy.ndarray
        The air temperature in degrees K
    rhumidity : float or numpy.ndarray
        Relative humidity, as a proportion on [0,1]
    vpd : float or numpy.ndarray
        The vapor pressure deficit in Pascals
    lai : float or numpy.ndarray
        The leaf area index (LAI)
    fpar : float or numpy.ndarray
        Fraction of photosynthetically active radiation (PAR) absorbed by
        the vegetation canopy
    rad_canopy : float or numpy.ndarray
        Net radiation to the canopy (J m-2 s-1)
    f_wet : float or numpy.ndarray or None
        (Optional) Fraction of the land surface that is saturated/ covered
        with standing water; if None, calculates this fraction like MOD16,
        based on the relative humidity
    g_h : float
        Leaf conductance to sensible heat per unit LAI
        (Default: 0.01 m s-1 LAI-1)
    g_e : float
        Leaf conductance to evaporated water per unit LAI
        (Default: 0.01 m s-1 LAI-1)

    Returns
    -------
    float or numpy.ndarray
        Evaporation from the wet canopy surface (kg m-2 s-1)
    '''
    assert np.logical_and(0 <= rhumidity, rhumidity <= 1),\
        'Relative humidity (rhumidity) must be on the interval [0,1]'
    if f_wet is None:
        f_wet = np.where(rhumidity < 0.7, 0, np.power(rhumidity, 4))
    # NIST simplified air density formula with buoyancy correction (NPL 2021)
    rho = np.divide( # Convert Pa to mbar, RH to RH% (percentage)
        0.348444 * (pressure / 100) - (rhumidity * 100) *\
            (0.00252 * temp_k - 273.15 - 0.020582),
        temp_k) # kg m-3
    # Wet canopy resistance to sensible heat ("rhc")
    r_h = 1 / (g_h * lai * f_wet)
    # Wet canopy resistance to evaporated water on the surface ("rvc")
    r_c = 1 / (g_e * lai * f_wet)
    # Resistance to radiative heat transfer through air ("rrc")
    r_r = np.divide(
        rho * SPECIFIC_HEAT_CAPACITY_AIR,
        4 * STEFAN_BOLTZMANN * np.power(temp_k, 3))
    # Aerodynamic resistance to evaporated water on the wet canopy surface
    r_a = np.divide(r_h * r_r, r_h + r_r) # (s m-1)
    # Slope of saturation vapor pressure curve
    s = svp_slope(temp_k)
    # Latent heat of vaporization (J kg-1)
    lhv = latent_heat_vapor(temp_k)
    # Mu et al. (2011), Equation 17; PET (J m-2 s-1) is divided by the latent
    #   heat of vaporization (J kg-1) to obtain mass flux (kg m-2 s-1)
    return np.divide(
        f_wet * ((s * rad_canopy) +\
            (rho * SPECIFIC_HEAT_CAPACITY_AIR * vpd * (fpar * 1/r_a))),
        s + ((pressure * SPECIFIC_HEAT_CAPACITY_AIR * r_c) *\
            1/(lhv * MOL_WEIGHT_WET_DRY_RATIO_AIR * r_a))) / lhv


def psychrometric_constant(pressure, temp_k):
    r'''
    The psychrometric constant, which relates the vapor pressure to the air
    temperature. Calculation derives from the "Handbook of Hydrology" by D.R.
    Maidment, Section 4.2.28.

    $$
    \gamma = \frac{C_p \times P}{\lambda\times 0.622}
    $$

    Where `C_p` is the specific heat capacity of air, `P` is air pressure, and
    `lambda` is the latent heat of vaporization.

    Parameters
    ----------
    pressure : float or numpy.ndarray
        The air pressure in Pascals
    temp_k : float or numpy.ndarray
        The air temperature in degrees K

    Returns
    -------
    float or numpy.ndarray
        The psychrometric constant at this pressure, temperature (Pa K-1)
    '''
    lhv = latent_heat_vapor(temp_k) # Latent heat of vaporization (J kg-1)
    return (SPECIFIC_HEAT_CAPACITY_AIR * pressure) /\
        (lhv * MOL_WEIGHT_WET_DRY_RATIO_AIR)


def radiation_net(sw_rad, sw_albedo, temp_k):
    r'''
    Net incoming radiation to the land surface, calculated according to the
    MOD16 algorithm and Cleugh et al. (2007); see Equation 7 in the MODIS
    MOD16 Collection 6.1 User's Guide.

    - Cleugh, H. A., Leuning, R., Mu, Q., & Running, S. W. (2007).
      Regional evaporation estimates from flux tower and MODIS satellite data.
      *Remote Sensing of Environment*, 106(3), 285–304.

    $$
    R_{net} = (1 - \alpha)\times R_{S\downarrow} +
        (\varepsilon_a - \varepsilon_s) \times \sigma \times T^4
    \quad\mbox{where}\quad \varepsilon_s = 0.97
    $$

    Where `alpha` is the MODIS albedo, `R_S` is down-welling short-wave
    radiation, `sigma` is the Stefan-Boltzmann constant, and:

    $$
    \varepsilon_a = 1 - 0.26\,\mathrm{exp}\left(
      -7.77\times 10^{-4}\times (T - 273.15)^2
    \right)
    $$

    Parameters
    ----------
    sw_rad : float or numpy.ndarray
        Incoming short-wave radiation (W m-2)
    sw_albedo : float or numpy.ndarray
        White-sky albedo, e.g., from MODIS MCD43A3
    temp_k : float or numpy.ndarray
        Air temperature in degrees K

    Returns
    -------
    float or numpy.ndarray
        Net incoming radiation to the land surface (W m-2)
    '''
    # Mu et al. (2011), Equation 5
    emis_surface = 0.97
    emis_atmos = 1 - 0.26 * np.exp(-7.77e-4 * np.power(temp_k - 273.15, 2))
    return sw_rad * (1 - sw_albedo) +\
        STEFAN_BOLTZMANN * (emis_atmos - emis_surface) * np.power(temp_k, 4)


def svp(temp_k):
    r'''
    The saturation vapor pressure, based on [
    the Food and Agriculture Organization's (FAO) formula, Equation 13
    ](http://www.fao.org/3/X0490E/x0490e07.htm).

    $$
    \mathrm{SVP} = 1\times 10^3\left(
    0.6108\,\mathrm{exp}\left(
      \frac{17.27 (T - 273.15)}{T - 273.15 + 237.3}
      \right)
    \right)
    $$

    Parameters
    ----------
    temp_k : float or numpy.ndarray
        The air temperature in degrees K

    Returns
    -------
    float or numpy.ndarray
    '''
    temp_c = temp_k - 273.15
    # And convert from kPa to Pa
    return 1e3 * 0.6108 * np.exp((17.27 * temp_c) / (temp_c + 237.3))


def svp_slope(temp_k):
    r'''
    The slope of the saturation vapour pressure curve, which describes the
    relationship between saturation vapor pressure and temperature. Based on [
    the Food and Agriculture Organization's (FAO) formula, Equation 13
    ](http://www.fao.org/3/X0490E/x0490e07.htm).

    $$
    \Delta = 4098\times [\mathrm{SVP}]\times (T - 273.15 + 237.3)^{-2}
    $$

    Parameters
    ----------
    temp_k : float or numpy.ndarray
        The air temperature in degrees K

    Returns
    -------
    float or numpy.ndarray
        The slope of the saturation vapor pressure curve in Pascals per
        degree K (Pa degK-1)
    '''
    temp_c = temp_k - 273.15
    svp = 0.6108 * np.exp((17.27 * temp_c) / (temp_c + 237.3))
    # With conversion of SVP from kPa to Pa
    return (4098 * (svp * 1e3)) / np.power(temp_c + 237.3, 2)


def transpiration_potential(
        air_temp_k, pressure, fpar, sw_rad, sw_albedo, rhumidity = None,
        f_wet = None):
    r'''
    Estimates potential transpiration in each soil layer. Transpiration is
    calculated using the Priestly-Taylor method for "potential transpiration,"
    as described in the MODIS MOD16 framework.

    $$
    \lambda T = \frac{\alpha\, s\, A_c\, (1 - F_{wet})}{\gamma\, s}
    $$

    Where `alpha` is the MODIS albedo, `s` is the slope of the saturation
    vapor pressure curve or `svp_slope()`, `A_c` is the net radiation
    intercepted by the canopy, `F_wet` is the fraction of the land surface
    that is saturated, and `gamma` is the `psychrometric_constant()`.

    Parameters
    ----------
    air_temp_k : numpy.ndarray
        Air temperature (deg K)
    pressure : numpy.ndarray
        Air pressure (Pa)
    fpar : numpy.ndarray
        Fraction of photosynthetically active radiation (PAR) absorbed by
        the vegetation canopy
    sw_rad : numpy.ndarray
        Incoming short-wave radiation, (W m-2) or (J s-1 m-2)
    sw_albedo : numpy.ndarray
        White-sky albedo for short-wave radiation
    rhumidity: numpy.ndarray or None
        (Optional) The relative humidity, expressed as a proportion on the
        interval [0,1]; if not provided, f_wet must be provided.
    f_wet : numpy.ndarray or None
        (Optional) Fraction of the land surface that is saturated/ covered
        with standing water; if None, calculates this fraction like MOD16,
        based on the relative humidity (rhumidity).

    Returns
    -------
    numpy.ndarray
        A (Z x 1) array of transpiration in each soil layer (kg m-2 s-1)
    '''
    assert rhumidity is not None or f_wet is not None,\
        'One must be provided: "rhumidity" or "f_wet"'
    if rhumidity is not None:
        assert np.logical_or(np.isnan(rhumidity),
                np.logical_and(0 <= rhumidity, rhumidity <= 1)).all(),\
            'Relative humidity (rhumidity) must be on the interval [0,1]'
    if f_wet is None:
        f_wet = np.where(rhumidity < 0.7, 0, np.power(rhumidity, 4))
    alpha = 1.26 # Equation 18, MOD16 Collection 6.1 User's Guide
    s = svp_slope(air_temp_k) # Slope of saturation vapor pressure curve
    gamma = psychrometric_constant(pressure, air_temp_k)
    # Net radiation to the land surface
    net_rad = radiation_net(sw_rad, sw_albedo, air_temp_k)
    # Net radiation to the canopy
    canopy_rad = np.multiply(fpar, net_rad)
    pt = np.divide(alpha * s * canopy_rad * (1 - f_wet), s + gamma)
    # Divide by latent heat of vaporization to convert to a mass flux
    return pt / latent_heat_vapor(air_temp_k)



'''
Functions for affine transformations between geographic coordinates (WGS84)
and EASE-Grid 2.0 row-column coordinates.

NOTE: The `ease2_from_wgs84()` and `ease2_to_wgs84()` functions should refer
to the center of a grid cell, which can be confirmed by:

    >>> ease2_from_wgs84(ease2_to_wgs84((0, 0), 'M09'), 'M01')
    (4, 4)
'''

import numpy as np
from functools import partial
from itertools import product
from pyl4c import haversine
from pyl4c.data.fixtures import EASE2_GRID_PARAMS

EARTH_RADIUS = 6378137.0 # WGS84 sphere
WGS84_ECCENTRICITY = 0.081819190843


def ease2_coords(grid: str, in_1d: bool = True):
    '''
    Returns EASE-Grid 2.0 coordinate arrays; coordinates are WGS84 longitude-
    or latitude.

    Parameters
    ----------
    grid : str
        The EASE-Grid 2.0 designation: M01, M09, etc.
    in_1d : bool
        True to produce 1D arrays (Default: True)

    Returns
    -------
    tuple
        Two elements, the longitude and latitude coordinate arrays:
        (numpy.ndarray, numpy.ndarray)
    '''
    nrows, ncols = EASE2_GRID_PARAMS[grid]['shape']
    rows = np.arange(0, nrows)
    cols = np.arange(0, ncols)
    e2ll = partial(ease2_to_wgs84, grid = grid)
    x_coords = np.array(list(map(lambda c: e2ll((0, c))[0], cols.tolist())))
    y_coords = np.array(list(map(lambda r: e2ll((r, 0))[1], rows.tolist())))
    if in_1d:
        return (x_coords, y_coords)
    return ( # Otherwise, return two 2D arrays with duplicates
        np.repeat(x_coords.reshape((1, ncols)), nrows, axis = 0),
        np.repeat(y_coords.reshape((nrows, 1)), ncols, axis = 1))


def ease2_nested_cells(coords, k: int, grid = 'M01'):
    '''
    Finds the row-column coordinates of all EASE-Grid 2.0 cells nested within
    a larger cell. The finer or target grid resolution is specified by `grid`;
    it is assumed that the WGS84 `coords` provided refer to the center of a
    larger cell.

    Parameters
    ----------
    coords : list or tuple
        The longitude-latitude pair for which to find the nearest grid cells
    k : int
        The search radius; corresponds to the number of rows and
            columns away from the defined coordinates
    grid : str
        The EASE-Grid 2.0 designation: M01, M09, etc.

    Returns
    -------
    list
    '''
    # Translate WGS84 to row-column indices
    row, col = ease2_from_wgs84(coords, grid)
    potential_rows = np.arange(row - k, row + k + 1, 1)
    potential_cols = np.arange(col - k, col + k + 1, 1)
    return list(product(potential_rows, potential_cols))


def ease2_search_radius(coords, k: int, grid = 'M09'):
    '''
    Finds the nearest EASE-Grid 2.0 grid cells using a search radius of k grid
    cells; returns a list of grid cell indices sorted by distance ascending.
    The row-column indices that correspond to the affine transformation of the
    given coordinates will always be the nearest, but this can be useful for
    finding the next-nearest row-column coordinates. See
    `pyl4c.ease2.ease2_nested_cells()`.

    Parameters
    ----------
    coords : list or tuple
        The longitude-latitude pair for which to find the nearest grid cells
    k : int
        The search radius; corresponds to the number of rows and
            columns away from the defined coordinates
    grid : str
        The EASE-Grid 2.0 designation: M01, M09, etc.

    Returns
    -------
    list
    '''
    choices_idx = ease2_nested_cells(coords, k, grid)
    choices_wgs84 = [ # Convert from row-column to longitude-latitude
        ease2_to_wgs84(pair, grid) for pair in choices_idx
    ]
    dists = [ # Calculate every pair-wise distance
        haversine(coords, choices_wgs84[i]) for i in range(0, len(choices_idx))
    ]
    # Return indices sorted by distance ascending
    return np.array(choices_idx)[np.argsort(dists),:].tolist()


def ease2_from_wgs84(coords, grid: str = 'M09', exact: bool = False):
    '''
    Given a longitude-latitude coordinate pair, derive EASE-Grid 2.0
    row-column coordinates. This is translated directly from the gridutil C
    code (easegrid_wgs84_transforms.c) which, in turn, was translated from
    Brodzik's IDL code. NOTE: Only supports the global EASE-Grid 2.0
    projection for now. NOTE: The `ease2_from_wgs84()` and `ease2_to_wgs84()`
    functions are only consistent (i.e., reversible) if `exact=True` is used.

    Parameters
    ----------
    coords : list or tuple
        Sequence of two floats, the longitude and latitude, respectively
    grid : str
        String, the EASE-Grid 2.0 designation: M01, M09, etc.
    exact : bool
        True to return grid cell indices as floating point instead of integer
        type (Default: False)

    Returns
    -------
    tuple
        The (row, column) coordinates corresponding to the given
        latitude-longitude coordinates
    '''
    # Unpack longitude and latitude
    user_lng, user_lat = coords

    # Define projection parameters for WGS84 and the EASE-Grid 2.0 projections
    e2 = np.power(WGS84_ECCENTRICITY, 2)
    epsilon = 1e-6

    # Unique to global EASE-Grid 2.0 projection
    ref_lat = ref_lng = 0.0
    ref_lat2 = 30.0
    sin_phi1 = np.sin(np.deg2rad(ref_lat2))
    cos_phi1 = np.cos(np.deg2rad(ref_lat2))
    kz = cos_phi1 / np.sqrt(1 - e2 * sin_phi1 * sin_phi1)

    # Get grid shape and resolution
    nrows, ncols = EASE2_GRID_PARAMS[grid]['shape']
    resolution = EASE2_GRID_PARAMS[grid]['resolution']
    r0 = (ncols - 1) / 2.0 # Column mapped to reference latitude
    s0 = (nrows - 1) / 2.0 # Row mapped to reference latitude(?)

    # Start transformation of user coordinates
    dlng = user_lng - ref_lng
    assert abs(dlng) <= 180, 'Longitude error'

    # Convert to radians
    phi = np.deg2rad(user_lat)
    lam = np.deg2rad(dlng)

    q = (1.0 - e2) * ((np.sin(phi) / (1.0 - e2 * np.sin(phi) * np.sin(phi))) - (1.0 / (2.0 * WGS84_ECCENTRICITY)) * np.log((1.0 - WGS84_ECCENTRICITY * np.sin(phi)) / (1.0 + WGS84_ECCENTRICITY * np.sin(phi))))
    qp = 1.0 - ((1.0 - e2) / (2.0 * WGS84_ECCENTRICITY) * np.log((1.0 - WGS84_ECCENTRICITY) / (1.0 + WGS84_ECCENTRICITY)))

    # Unique to global EASE-Grid 2.0 projection
    x = EARTH_RADIUS * kz * lam
    y = EARTH_RADIUS * (q / (2.0 * kz))

    result = (s0 - (y / resolution), r0 + (x / resolution))
    if exact:
        return result

    return tuple(map(lambda c: int(round(c)), result))


def ease2_to_wgs84(coords, grid = 'M09'):
    '''
    Given a row-column coordinate pair from an EASE-Grid 2.0, derive the
    corresponding WGS84 longitude-latitude coordinates. This is translated
    directly from the gridutil C code (easegrid_wgs84_transforms.c) which, in
    turn, was translated from Brodzik's IDL code. NOTE: Only supports the
    global EASE-Grid 2.0 projection for now. NOTE: The `ease2_from_wgs84()`
    and `ease2_to_wgs84()` functions are only consistent (i.e., reversible) if
    `exact=True` is used (in `ease2_from_wgs84()`).

    Parameters
    ----------
    coords : list or tuple
        Sequence of two integers, the row and column coordinates
    grid : str
        The EASE-Grid 2.0 designation: M01, M09, etc.

    Returns
    -------
    tuple
        The (longitude, latitude) coordinates corresponding to the given
        row-column coordinates
    '''
    e2 = np.power(WGS84_ECCENTRICITY, 2)
    e4 = np.power(WGS84_ECCENTRICITY, 4)
    e6 = np.power(WGS84_ECCENTRICITY, 6)
    epsilon = 1e-6

    # Unique to global EASE-Grid 2.0 projection
    ref_lat = ref_lng = 0.0
    ref_lat2 = 30.0
    sin_phi1 = np.sin(np.deg2rad(ref_lat2))
    cos_phi1 = np.cos(np.deg2rad(ref_lat2))
    kz = cos_phi1 / np.sqrt(1 - e2 * sin_phi1 * sin_phi1)

    # Convert (row, column) to (x, y) coordinates for this projection
    x, y = translate_row_col_to_ease2(coords, grid)

    qp = (1.0 - e2) * ((1.0 / (1.0 - e2)) - (1.0 / (2.0 * WGS84_ECCENTRICITY)) * np.log((1.0 - WGS84_ECCENTRICITY) / (1.0 + WGS84_ECCENTRICITY)))

    # Unique to global EASE-Grid 2.0 projection
    beta = np.arcsin(2.0 * y * kz / (EARTH_RADIUS * qp))
    lam  = x / (EARTH_RADIUS * kz)

    phi = beta \
        + (((e2 / 3.0) + ((31.0 / 180.0) * e4) \
              + ((517.0 / 5040.0) * e6)) * np.sin(2.0 * beta)) \
        + ((((23.0 / 360.0) * e4) \
              + ((251.0 / 3780.0) * e6)) * np.sin(4.0 * beta)) \
        + (((761.0 / 45360.0) * e6) * np.sin(6.0 * beta))

    return ((180.0 * (lam / np.pi)) + ref_lng, (180.0 * phi / np.pi))


def translate_row_col_to_ease2(coords, grid = 'M09'):
    '''
    Returns the EASE-Grid 2.0 coordinates at the given row-column index.

    Parameters
    ----------
    coords : list or tuple
        Sequence of two integers, the row and column coordinates
    grid : str
        The EASE-Grid 2.0 designation: M01, M09, etc.

    Returns
    -------
    tuple
        The EASE-Grid 2.0 (X, Y) coordinates corresponding to the given
        row-column coordinates
    '''
    # Unpack row, column coordinates
    user_row, user_col = coords

    # Get grid shape and resolution
    nrows, ncols = EASE2_GRID_PARAMS[grid]['shape']
    resolution = EASE2_GRID_PARAMS[grid]['resolution']
    r0 = (ncols - 1) / 2.0 # Column mapped to reference latitude
    s0 = (nrows - 1) / 2.0 # Row mapped to reference latitude(?)

    # Convert (row, column) to (x, y) coordinates for this projection
    x = (user_col - r0) * resolution
    y = (s0 - user_row) * resolution
    return (x, y)



'''
Helper functions and data (EASE-Grid 2.0 parameters) for converting SMAP L4C
data from its native HDF5 format into or out of various GIS-friendly formats.
It may seem arbitrary that the subset() function is in the `utils` module and
not here, but the main distinction is that `spatial` module includes functions
that depend on the GDAL library (osgeo.gdal) but `utils` does not and should
not.

Corner coordinates for EASE-Grid 2.0 obtained from [1] and EPSG codes
obtained from [2]. The same `proj4` string is used for all EASE2 grids; see [3].

    # proj4 string
    +proj=cea +lat_0=0 +lon_0=0 +lat_ts=30 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units=m

Note also that `array_to_raster()`, `array_to_raster_clone()`, `as_array()`,
and `dump_raster()` are copied from my (MIT-licensed) `unmixing` library [4].

1. https://nsidc.org/ease/ease-grid-projection-gt
2. https://nsidc.org/data/ease
3. https://doi.org/10.3390/ijgi3031154
4. https://github.com/arthur-e/unmixing/
'''

import tempfile
import numpy as np
import pyproj
from osgeo import gdal, gdalconst, gdal_array, osr
from affine import Affine
from pyl4c.data.fixtures import EASE2_GRID_PARAMS
from pyl4c.epsg import EPSG
from pyl4c.utils import sample

def array_to_raster(
        a, gt, wkt, xoff = None, yoff = None, dtype = None, nodata = -9999):
    '''
    Creates a raster from a given array, with optional x- and y-offsets
    if the array was clipped.

    Parameters
    ----------
    a : numpy.ndarray
        NumPy array to convert to a raster dataset
    gt : tuple
        A GDAL GeoTransform tuple
    wkt : str
        Spatial reference system (SRS) as Well-Known Text projection string
    xoff : int
        The offset in the x-direction; should be provided when clipped
    yoff : int
        The offset in the y-direction; should be provided when clipped
    dtype : type
        The data type to coerce on the array
    nodata : int or float
        The NoData or Fill value (Default: -9999)

    Returns
    -------
    gdal.Dataset
    '''
    if dtype is not None:
        a = a.astype(dtype)
    try:
        rast = gdal_array.OpenNumPyArray(a)
    except AttributeError:
        # For backwards compatibility with older version of GDAL
        rast = gdal.Open(gdal_array.GetArrayFilename(a))
    except:
        rast = gdal_array.OpenArray(a)
    rast.SetGeoTransform(gt)
    rast.SetProjection(wkt)
    if nodata is not None:
        for band in range(1, rast.RasterCount + 1):
            rast.GetRasterBand(band).SetNoDataValue(nodata)
    if xoff is not None and yoff is not None:
        # Bit of a hack; essentially, re-create the raster but with the
        #   correct X and Y offsets (don't know how to do this without the
        #   use of CopyDatasetInfo())
        return array_to_raster_clone(a, rast, xoff, yoff)
    return rast


def array_to_raster_clone(a, proto, xoff = None, yoff = None):
    '''
    Creates a raster from a given array and a prototype raster dataset, with
    optional x- and y-offsets if the array was clipped.

    Parameters
    ----------
    a : numpy.ndarray
        NumPy array to convert to a raster dataset
    proto : gdal.Dataset
        A prototype dataset
    xoff : int or None
        The offset in the x-direction; should be provided when clipped
    yoff : int or None
        The offset in the y-direction; should be provided when clipped

    Returns
    -------
    gdal.Dataset
    '''
    try:
        rast = gdal_array.OpenNumPyArray(a)
    except AttributeError:
        # For backwards compatibility with older version of GDAL
        rast = gdal.Open(gdal_array.GetArrayFilename(a))
    except:
        rast = gdal_array.OpenArray(a)
    kwargs = dict()
    if xoff is not None and yoff is not None:
        kwargs = dict(xoff=xoff, yoff=yoff)
    # Copy the projection info and metadata from a prototype dataset
    if type(proto) == str:
        proto = gdal.Open(proto)
    gdal_array.CopyDatasetInfo(proto, rast, **kwargs)
    return rast


def as_array(file_path, band_axis = True, subset_idx = None, nodata = None):
    '''
    A convenience function for opening a raster as an array and accessing its
    spatial information; takes a single string argument.

    NOTE: subset_idx should be provided as either a flat sequence or nested
    tuples. For a flat sequence (GDAL format):

        (<xoff>, <yoff>, <xmax>, <ymax>)

    Where `<xoff>, <yoff>` are the offsets in the X- and Y-directions (column
    and row directions, starting from the top-left corner. The last two
    numbers indicate the ending column and row indices.

    For nested tuples (for backwards compatibility):

        ((<xoff>, <xmax>), (<yoff>, <ymax>))

    For instance, if the offsets equal zero and the `<xmax>, <ymax>` terms are
    equal to the total number of columns and rows, respectively, then the
    entire raster array is read (i.e., might as well set `subset_idx = None`).

    Parameters
    ----------
    file_path : str
        The path of the raster file to open as an array
    band_axis : bool
        True to include a band axis, even for single-band rasters
    subset_idx : tuple or list
        (Optional) Can provide any arbitrary bounding box sequence as
        row-column indices (not geographic coordinates) to get a spatial
        subset instead of the entire raster array.
    nodata : int or float
        (Optional) If not None, fills this value with NaN (numpy.nan)

    Returns
    -------
    tuple
        Sequence of (numpy.ndarray, geotransform tuple, projection str)
    '''
    ds = gdal.Open(file_path)
    if len(ds.GetSubDatasets()) > 0:
        raise ValueError(
            'This dataset has multiple layers; you must specify a valid subdataset, for example: %s' % ds.GetSubDatasets()[0][0])
    if subset_idx is not None:
        bb = subset_idx
        # If nested sequence is passed
        if hasattr(subset_idx[0], '__len__'):
            bb = [bb[0][0], bb[1][0], bb[0][1], bb[1][1]]
        # Default is for GDAL-style bounds (<xmin>, <ymin>, <xmax>, <ymax>)
        assert all(map(lambda x: x >= 0, bb)),\
            'Elements of subset_idx must be slice indices >/= 0'
        arr = ds.ReadAsArray(
            *map(int, (bb[0], bb[1], bb[2] - bb[0], bb[3] - bb[1])))
    else:
        arr = ds.ReadAsArray()
    gt = ds.GetGeoTransform()
    wkt = ds.GetProjection()
    ds = None
    # Make sure that single-band rasters have a band axis
    if band_axis and len(arr.shape) < 3:
        shp = arr.shape
        arr = arr.reshape((1, shp[0], shp[1]))
    if nodata is not None:
        arr[arr == nodata] = np.nan
    return (arr, gt, wkt)


def as_raster(file_path, subset_bbox = None):
    '''
    A convenience function for opening a raster and accessing its spatial
    information; takes a single string argument.

    NOTE: subset_bbox should be provided as either a flat sequence or nested
    tuples. For a flat sequence (GDAL format):

        (<xmin>, <ymin>, <xmax>, <ymax>)

    For nested tuples (for backwards compatibility):

        ((<xmin>, <xmax>), (<ymin>, <ymax>))

    Parameters
    ----------
    file_path : str
        The path of the raster file to open as a gdal.Dataset
    subset_bbox : tuple or list
        (Optional) The bounding-box coordinates, as geographic coordinate
        pairs, to use in extracting a spatial subset

    Returns
    -------
    tuple
        Tuple of: (gdal.Dataset, geotransform tuple, projection str)
    '''
    ds = gdal.Open(file_path)
    gt = ds.GetGeoTransform()
    wkt = ds.GetProjection()
    if subset_bbox is not None:
        bb = subset_bbox
        # If nested sequence is passed
        if hasattr(subset_bbox[0], '__len__'):
            bb = [bb[0][0], bb[1][0], bb[0][1], bb[1][1]]

        # NOTE: Because SRS coordinate system is "flipped" vertically
        #   w.r.t. pixel coordinate system, we flip ymax and ymin;
        # See: https://www.perrygeo.com/python-affine-transforms.html
        ymax, xmin = xy_to_pixel(bb[0:2], gt)
        ymin, xmax = xy_to_pixel(bb[2:4], gt)
        array, gt, wkt = as_array( # Read in as array
            file_path, band_axis, (xmin, ymin, xmax, ymax))
        ds = array_to_raster(array, gt, wkt, xmin, ymin)
        # Get updated GeoTransform, with correct upper-left coordinate pair
        gt = ds.GetGeoTransform()
    return (ds, gt, wkt)


def binary_reclass(a, target_input = (0,), output = (0, 1)):
    '''
    For an input raster array, produce a binary output array, replacing all
    cases of `target_input` values with outputs[0] and all other cases with
    output[1].

    Parameters
    ----------
    a : numpy.ndarray
        A NumPy array to reclassify
    target_input : tuple or list or numpy.ndarray
        Vector of input values to replace with output[0]
    output : tuple or list or numpy.ndarray
        2-element vector of values which will replace input values

    Returns
    -------
    numpy.ndarray
    '''
    assert len(output) == 2, 'Only two output values should be provided'
    return np.where(np.isin(a, target_input), output[0], output[1])


def bounds(rast):
    '''
    Returns the minimum and maximum coordinate values in the sequence expected
    by, e.g., the `-te` switch in various GDAL utiltiies:
    `(xmin, ymin, xmax, ymax)`. Copied from my MIT-licensed library
    `gdal_extent.py`.

    Parameters
    ----------
    rast : gdal.Dataset

    Returns
    -------
    tuple
        The bounds of the raster, i.e., `(xmin, ymin, xmax, ymax)`
    '''
    gt = rast.GetGeoTransform()
    xsize = rast.RasterXSize # Size in the x-direction
    ysize = rast.RasterYSize # Size in the y-direction
    xr = abs(gt[1]) # Resolution in the x-direction
    yr = abs(gt[-1]) # Resolution in the y-direction
    return (gt[0], gt[3] - (ysize * yr), gt[0] + (xsize * xr), gt[3])


def crs_transform(from_epsg, to_epsg):
    '''
    Creates a coordinate reference system (CRS) transformation. The resulting
    `CoordinateTransformation` instance should be used, e.g.:

        crs_transform(4326, 6933).TransformPoint(latitude, longitude)

    Parameters
    ----------
    from_epsg : int
        The EPSG code for the originating CRS
    to_epsg : int
        The EPSG code for the target (output) CRS

    Returns
    -------
    gdal.osr.CoordinateTransformation
    '''
    source = osr.SpatialReference()
    target = osr.SpatialReference()
    source.ImportFromEPSG(from_epsg)
    target.ImportFromEPSG(to_epsg)
    return osr.CoordinateTransformation(source, target)


def dump_raster(
        rast, rast_path, driver = 'GTiff', gdt = None, nodata = None,
        compress = False):
    '''
    Creates a raster file from a given `gdal.Dataset` instance.

    Parameters
    ----------
    rast : gdal.Dataset
        A gdal.Dataset; does NOT accept NumPy array
    rast_path : str
        The path of the output raster file
    driver : str
        The name of the GDAL driver to use, which determines output file type
        (Default: "GTiff")
    gdt : int
        The GDAL data type to use, e.g., see gdal.GDT_Float32
    nodata : int or float
        The NoData value; defaults to -9999.
    compress : bool
        True to apply lossless compression to the output; requires that the
        "GTiff" driver is used (GeoTIFF output)
    '''
    if gdt is None:
        gdt = rast.GetRasterBand(1).DataType
    driver = gdal.GetDriverByName(driver)
    if compress and driver.ShortName == 'GTiff':
        _rast_path = rast_path
        tmp = tempfile.NamedTemporaryFile()
        rast_path = tmp.name
    sink = driver.Create(
        rast_path, rast.RasterXSize, rast.RasterYSize, rast.RasterCount, int(gdt))
    assert sink is not None,\
        'Cannot create dataset; there may be a problem with the output path you specified'
    sink.SetGeoTransform(rast.GetGeoTransform())
    sink.SetProjection(rast.GetProjection())
    for b in range(1, rast.RasterCount + 1):
        dat = rast.GetRasterBand(b).ReadAsArray()
        sink.GetRasterBand(b).WriteArray(dat)
        sink.GetRasterBand(b).SetStatistics(*map(np.float64,
            [dat.min(), dat.max(), dat.mean(), dat.std()]))
        if nodata is None:
            nodata = rast.GetRasterBand(b).GetNoDataValue()
            if nodata is None:
                nodata = -9999
        sink.GetRasterBand(b).SetNoDataValue(np.float64(nodata))
    sink.FlushCache()
    if not compress or driver.ShortName != 'GTiff':
        return # Done
    # Optionally, compress the output file
    opts = gdal.TranslateOptions(
        format = 'GTiff', creationOptions = ['COMPRESS=LZW'])
    # Note that "_rast_path" is the true output path, "rast_path" just temp
    gdal.Translate(_rast_path, rast_path, options = opts)


def ease2_coords_approx(grid, in_1d = True, srs_epsg = 4326, precision = 8):
    '''
    Returns approximate EASE-Grid 2.0 coordinate arrays. They are
    "approximate" because, although they are calculated correctly using affine
    transformations, they do not correspond to the coordinate arrays provided
    in typical EASE-Grid 2.0 products, namely SMAP L4C. They only correspond
    with these arrays to 2nd or 3rd decimal place in geographic coordinates
    (i.e., with `srs_epsg = 4326`).

    Parameters
    ----------
    grid : str
        Output grid size: "M01" or "M09"
    in_1d : bool
        True to produce 1D arrays (Default: True)
    srs_epsg : int
        The EPSG code of the desired output SRS (Default: 4326)
    precision : int
        The decimal precision of output coordinates (Default: 8)

    Returns
    -------
    tuple
        Two coordinate arrays, X and Y: (numpy.ndarray, numpy.ndarray)
    '''
    srs0 = osr.SpatialReference()
    srs0.ImportFromWkt(EPSG[EASE2_GRID_PARAMS[grid]['epsg']])
    srs = osr.SpatialReference()
    srs.ImportFromEPSG(srs_epsg)
    native = Affine.from_gdal(*EASE2_GRID_PARAMS[grid]['geotransform'])
    target = osr.CoordinateTransformation(srs0, srs)
    shp = EASE2_GRID_PARAMS[grid]['shape']
    rows = np.arange(0, shp[0]) + 0.5 # To get center of a pixel
    cols = np.arange(0, shp[1]) + 0.5
    l1, l2 = (1, 0) if gdal.VersionInfo()[0] == '3' else (0, 1)
    x_coords = np.array([
        tuple(map( # Convert from (row, column) to (easting, northing)...
            # Then to (latitude, longitude)
            lambda x: round(x, precision),
            target.TransformPoint(*np.array(native * (i, 0)))))[l1]
        for i in cols
    ])
    y_coords = np.array([
        tuple(map(
            lambda x: round(x, precision),
            target.TransformPoint(*np.array(native * (0, i)))))[l2]
        for i in rows
    ])
    if in_1d:
        return (x_coords, y_coords)
    return (
        np.repeat(x_coords.reshape((1, shp[1])), shp[0], axis = 0),
        np.repeat(y_coords.reshape((shp[0], 1)), shp[1], axis = 1))


def ease2_to_geotiff(
        array, output_path, grid, xoff = None, yoff = None, dtype = None,
        nodata = -9999, compress = False):
    '''
    Convenience function for writing an EASE-Grid 2.0 data array to a
    GeoTIFF raster file.

    Parameters
    ----------
    array : numpy.ndarray
        A NumPy data array
    output_path : str
        The output GeoTIFF raster file path
    grid : str
        One of: "M09", "M01"
    xoff : int
        The offset in the x-direction; should be provided when clipped
    yoff : int
        The offset in the y-direction; should be provided when clipped
    dtype : type
        The NumPy data type to enforce, e.g., np.int32
    nodata : int or float
        The NoData or Fill value
    compress : bool
        True to apply lossless compression to the output (Default: False)
    '''
    # Check that the provided array size/shape makes sense
    expected = EASE2_GRID_PARAMS[grid]['shape']
    if xoff is None and yoff is None:
        assert (array.ndim == 2 and expected == array.shape) or (array.ndim == 3 and expected == array.shape[1:3]),\
        'Array shape (%d, %d) does not matched expected shape (%d, %d); set xoff and yoff if this array is clipped' % (*array.shape[1:3], *expected)
    gt = EASE2_GRID_PARAMS[grid]['geotransform']
    wkt = EPSG[EASE2_GRID_PARAMS[grid]['epsg']]
    rast = array_to_raster(array, gt, wkt, xoff, yoff, dtype = dtype)
    dump_raster(
        rast, output_path, driver = 'GTiff', nodata = nodata,
        compress = compress)


def ease2_hdf_to_geotiff(
        hdf, field, output_path, grid = 'M09', dtype = None, nodata = None):
    '''
    Dumps a single field from an HDF5 file into a GeoTIFF file. The grid
    argument refers to which EASE-Grid 2.0 is used:

    - M01: 1-km Global grid
    - M09: 9-km Global grid

    Parameters
    ----------
    hdf : h5py.File
        The HDF5 file / h5py File object
    field : str
        The hierarchical path to the variable of interest
    output_path : str
        The output file path
    grid : str
        EASE-Grid 2.0 type: "M01", "M09", ...
    dtype : type
        The NumPy data type to coerce on the array
    nodata : int or float
        The NoData value or "fill" value
    '''
    gt = EASE2_GRID_PARAMS[grid]['geotransform']
    wkt = EPSG[EASE2_GRID_PARAMS[grid]['epsg']]
    # Get the NoData value from the dataset description, if possible
    if nodata is None and '_FillValue' in hdf[field].attrs:
        nodata = hdf[field].attrs['_FillValue']
    ease2_to_geotiff(hdf[field][:], output_path, grid, dtype, nodata)


def intersect_rasters(
        ref_raster, src_raster, nodata = -9999, nudge = (0, 0),
        method = gdalconst.GRA_Bilinear, gdt = gdalconst.GDT_Float32):
    '''
    Projects the source raster so that its top-left corner is aligned with
    the reference raster. Then, clips or pads the source raster so that it
    has the same number of rows and columns (covers the same extent at the
    same grid resolution) as the reference raster. Can also "nudge" a raster
    in one or both directions, which can help fix small differences in
    alignment.

    NOTE: If the reference raster's top-left corner is far left and/or above
    that of the source raster, the intersected raster may contain no data
    from the original raster, i.e., an empty raster will result.

    NOTE: If the `ref_raster` and `src_raster` have different projections, the
    result may contain no data (i.e., all `NoData`). Not sure why, as here we
    are projecting the data to match the reference raster. This method should
    be used with rasters that are "close" but not an exact match for SRS,
    resolution, and extent.

    Parameters
    ----------
    ref_raster : gdal.Dataset
        The reference raster
    src_raster : gdal.Dataset
        The source raster; the raster to be changed
    nodata : int or float
        The NoData value to fill in where the reference raster is larger
    nudge : tuple or list
        Sequence of two integers, the number of pixels to push a raster array
        in the X and Y directions, respectively
    method : int
        gdal.GRA_Bilinear or gdal.GRA_NearestNeighbour are recommended
        (Default: gdal.GRA_Bilinear)
    gdt : int
        The GDAL data type to use for the output raster
        (Default: gdalconst.GDT_Float32)

    Returns
    -------
    gdal.Dataset
    '''
    msg = '%s is not a gdal.Dataset!'
    assert hasattr(ref_raster, 'GetGeoTransform'), msg % '"ref_raster"'
    assert hasattr(src_raster, 'GetGeoTransform'), msg % '"src_raster"'
    # Check that raster has spatial projection information
    msg = '%s has no defined spatial reference!'
    assert ref_raster.GetProjection() != '', msg % '"ref_raster"'
    assert src_raster.GetProjection() != '', msg % '"src_raster"'

    gt = ref_raster.GetGeoTransform()
    wkt = ref_raster.GetProjection()
    gt0 = src_raster.GetGeoTransform()
    wkt0 = src_raster.GetProjection()

    # Create a new raster with the desired attributes
    width, height = (ref_raster.RasterXSize, ref_raster.RasterYSize)
    width0, height0 = (src_raster.RasterXSize, src_raster.RasterYSize)
    out_raster = gdal.GetDriverByName('MEM').Create('temp.file',
        width, height, src_raster.RasterCount, gdt)

    # Initialize and set the NoData value
    for i in range(1, src_raster.RasterCount + 1):
        b = out_raster.GetRasterBand(i)
        b.Fill(nodata)
        b.SetNoDataValue(nodata)

    # Set the desired geotransform, and projection
    out_raster.SetGeoTransform(gt)
    out_raster.SetProjection(wkt)

    # Re-project the source image; now the top-left corners are aligned
    gdal.ReprojectImage(src_raster, out_raster, wkt0, wkt, method)
    arr0 = out_raster.ReadAsArray()
    del src_raster # Delete original raster references
    del ref_raster
    del out_raster
    # Clip the extent of the src image if ref is smaller
    if arr0.ndim > 2: ch, cw = arr0.shape[1:]
    else: ch, cw = arr0.shape
    if (width <= width0): cw = width # Clip src to ref extents
    if (height <= height0): ch = height

    # Read rows, columns out to the extent of the output array
    if arr0.ndim > 2:
        arr = arr0[:,0:ch,0:cw]
    else:
        arr = arr0[0:ch,0:cw]

    # Finally, "nudge" rows and columns over and down, as needed
    nx, ny = nudge
    # Nudge rows in the y-direction
    if ny > 0:
        arr = np.vstack((np.ones(arr.shape)[:ny,:] * np.nan, arr))
    elif ny < 0:
        arr = np.vstack((arr, np.ones(arr.shape)[:abs(ny),:] * np.nan))

    # Nudge columns in the x-direction
    if nx > 0:
        arr = np.hstack((np.ones(arr.shape)[:,:nx] * np.nan, arr))
    elif ny < 0:
        arr = np.hstack((arr, np.ones(arr.shape)[:,:abs(nx)] * np.nan))

    return array_to_raster(arr, gt, wkt)


def intersect_rasters2(
        ref_raster, src_raster, nodata = -9999,
        method = gdalconst.GRA_Bilinear, gdt = gdalconst.GDT_Float32):
    '''
    Improvement upon `intersect_rasters()`, perhaps? This function is
    considerably shorter as it uses a wrapper for the command-line `gdalwarp`.
    Will need to test performance.

    NOTE: If the reference raster's top-left corner is far left and/or above
    that of the source raster, the intersected raster may contain no data
    from the original raster, i.e., an empty raster will result.

    NOTE: If the `ref_raster` and `src_raster` have different projections, the
    result may contain no data (i.e., all `NoData`). Not sure why, as here we
    are projecting the data to match the reference raster.

    Parameters
    ----------
    ref_raster : gdal.Dataset
        The reference raster
    src_raster : gdal.Dataset
        The source raster; the raster to be changed
    nodata : int or float
        The NoData value to fill in where the reference raster is larger
    method : int
        gdal.GRA_Bilinear or gdal.GRA_NearestNeighbour are recommended
        (Default: gdal.GRA_Bilinear)
    gdt : int
        The GDAL data type to use for the output raster
        (Default: gdalconst.GDT_Float32)

    Returns
    -------
    gdal.Dataset
    '''
    msg = '%s is not a gdal.Dataset!'
    assert hasattr(ref_raster, 'GetGeoTransform'), msg % '"ref_raster"'
    assert hasattr(src_raster, 'GetGeoTransform'), msg % '"src_raster"'
    # Check that raster has spatial projection information
    msg = '%s has no defined spatial reference!'
    assert ref_raster.GetProjection() != '', msg % '"ref_raster"'
    assert src_raster.GetProjection() != '', msg % '"src_raster"'
    gt = ref_raster.GetGeoTransform()
    x_res = y_res = np.abs(gt[1])
    wkt_proj4 = pyproj.CRS.from_wkt(ref_raster.GetProjectionRef()).to_proj4()
    return gdal.Warp(
        'temp.file', src_raster, format = 'MEM', xRes = x_res, yRes = y_res,
        dstSRS = wkt_proj4, outputBounds = bounds(ref_raster))


def project_equirectangular(
        src_raster, size_degrees, method = gdal.GRA_Bilinear, nodata = -9999,
        gdt = gdalconst.GDT_Float32, nrows = None, ncols = None):
    '''
    Projects a raster into an equirectangular geographic coordinate system
    (GCS) based on the WGS84 datum. Thanks to J. Gómez-Dans for the more
    general solution to calculating output raster size [1].

    NOTE: This will not work well for very small cell sizes, around 0.1
    degress on a side or less. It should generally be used for down-sampling
    (scaling up) only.

    1. https://jgomezdans.github.io/gdal_notes/reprojection.html

    Parameters
    ----------
    src_raster : gdal.Dataset
        Raster to be projected to a GCS
    size_degrees : int or float
        Output cell size (spatial resolution) in degrees
    method : int
        gdal.GRA_Bilinear or gdal.GRA_NearestNeighbour are recommended
        (Default: gdal.GRA_Bilinear)
    gdt : int
        Output GDAL data type (Default: 32-bit floating point)

    Returns
    -------
    gdal.Dataset
    '''
    assert src_raster.RasterCount == 1, 'Can only work with 1-band raster arrays'
    gt0 = src_raster.GetGeoTransform()
    wkt0 = osr.SpatialReference()
    wkt0.ImportFromWkt(src_raster.GetProjection())

    # Equirectangular (target) spatial reference system
    wkt = osr.SpatialReference()
    wkt.ImportFromEPSG(4326)

    # NOTE: God forbid, if this transformation stops working, note that as of
    #   GDAL 3.0+, TransformPoint() expects coordinates to be provided in
    #   the order that the CRS expects, which is not necessarily X,Y,Z order
    transform = osr.CoordinateTransformation(wkt0, wkt)
    # Work out the boundaries of the new dataset in the target projection
    (uly, ulx, ulz) = transform.TransformPoint(gt0[0], gt0[3])
    (lry, lrx, lrz) = transform.TransformPoint(
        gt0[0] + gt0[1]*src_raster.RasterXSize,
        gt0[3] + gt0[5]*src_raster.RasterYSize)
    # Prior to GDAL 3.x, WGS84 projection defined coordinates in (X,Y), i.e.,
    #   (longitude, latitude) order; now they are (latitude, longitude) order
    if gdal.__version__[0] == '2':
        # So, for older versions, just swap coordinate order
        uly, ulx = (ulx, uly)
        lry, lrx = (lrx, lry)

    if ncols is None:
        ncols = abs(int((lrx - ulx) / size_degrees)) # Get NEW raster width
        # For lat-long projections, this is a sign there may be a problem
        if wkt.IsGeographic() and ncols <= 1:
            # Need to consider West longitude a possibility
            ncols = abs(int((-lrx - ulx) / size_degrees))
    if nrows is None:
        # Neat math fact: Not an issue for southern latitude
        nrows = int(abs((uly - lry) / size_degrees)) # Get NEW raster height
    assert ncols != 0 and nrows != 0,\
        'Rows and/or columns are zero; check that size_degrees is appropriate'

    gt = gdal.AutoCreateWarpedVRT(src_raster, str(wkt0), str(wkt), method)\
        .GetGeoTransform()
    gt = (gt[0], size_degrees, gt[2], gt[3], gt[4], -size_degrees)
    # Alternatively, we can guess some of the parameters (geog. CRS example):
    # gt = (-ulx, size_degrees, gt0[2], uly, gt0[4], -size_degrees)
    rast_projected = gdal.GetDriverByName('MEM')\
        .Create('', ncols, nrows, 1, gdt) # 1 refers to the number of bands

    # Set the NoData value so this is ignored in reprojecting
    src_raster.GetRasterBand(1).SetNoDataValue(nodata)
    output_band = rast_projected.GetRasterBand(1)
    output_band.SetNoDataValue(nodata)
    output_band.Fill(nodata) # This is necessary to place NoData in the output

    # Set the output image SRS and warp warp!
    rast_projected.SetGeoTransform(gt)
    rast_projected.SetProjection(str(wkt))
    gdal.ReprojectImage(
        src_raster, rast_projected, str(wkt0), str(wkt), method)
    return rast_projected


def resample_ease2(
        src_raster, grid, nodata = -9999, method = gdalconst.GRA_Bilinear,
        gdt = gdalconst.GDT_Float32):
    '''
    Resamples an input raster from one nested EASE-Grid 2.0 to another; e.g.,
    from a 1-km grid to a 9-km grid or vice-versa. Uses GDAL for resampling,
    which is expected to produce the same result as a matrix operations that
    exploit grid nesting.

    Parameters
    ----------
    src_raster : gdal.Dataset
        The source raster to be resampled
    grid : str
        Desired output EASE-Grid 2.0 specification, e.g., "M09"
    nodata : int or float
        The NoData value to fill in where the reference raster is larger
    method : int
        gdal.GRA_Bilinear or gdal.GRA_NearestNeighbour are recommended
        (Default: gdal.GRA_Bilinear)
    gdt : int
        The GDAL data type to use for the output raster
        (Default: gdalconst.GDT_Float32)

    Returns
    -------
    gdal.Dataset
    '''
    assert hasattr(src_raster, 'GetGeoTransform'), '"src_raster" is not a gdal.Dataset!'
    assert src_raster.GetProjection() != '', '"src_raster" has no defined spatial reference!'
    # Get attributes of the original (input) raster
    gt0 = src_raster.GetGeoTransform()
    wkt0 = src_raster.GetProjection()
    # Get attributes of the output raster
    res = EASE2_GRID_PARAMS[grid]['geotransform'][1]
    height, width = EASE2_GRID_PARAMS[grid]['shape']
    gt = list(gt0)
    gt[1] = res
    gt[-1] = -res
    # Create the output raster as an in-memory raster file
    out_raster = gdal.GetDriverByName('MEM').Create('temp.file',
        width, height, src_raster.RasterCount, gdt)
    # Set the desired geotransform, and projection
    out_raster.SetGeoTransform(tuple(gt))
    out_raster.SetProjection(wkt0)
    # We're using this empty raster as a reference for projecting the source
    return intersect_rasters(out_raster, src_raster, nodata, method, gdt)


def sample_points(data, points, gt = None, ignore_srs_mismatch = False):
    '''
    Samples the value in a raster array at each location in a sequence of
    geographic coordinate pairs.

    Parameters
    ----------
    data : numpy.ndarray or gdal.Dataset
    points : tuple or list
        Sequence of (X, Y) coordinate pairs
    gt : tuple
        (Optional) GeoTransform tuple
    ignore_srs_mismatch : bool
        True to ignore any apparent differences in the SRS between "points"
        and raster GeoTransform

    Returns
    -------
    list
        A list of sampled raster values
    '''
    if hasattr(data, 'GetGeoTransform'):
        if gt is not None and not np.all(np.equal(gt, rast.GetGeoTransform())):
            raise ValueError('Input raster GeoTransform and user-provided "gt" argument do not match (should drop "gt" argument)')
        gt = rast.GetGeoTransform()
        transform = Affine.from_gdal(*gt)
    else:
        if gt is None:
            raise ValueError('Must specify a geotransform, gt, unless a gdal.Dataset is provided')
        transform = Affine.from_gdal(*gt)

    # Check that the point X coordinate is not too many orders of magnitude
    #   different from the minimum X coordinate of the input raster SRS; if
    #   so, there is likely a mismatch in SRS
    if abs(np.log10(abs(gt[0])) - np.log10(abs(points[0][0]))) >= 2:
        if not ignore_srs_mismatch:
            raise ValueError('Coordinate system of "points" and input data do not match')

    # If points are given as a sequence of pairs, transformation to
    #   row-column indices is straightforward
    if hasattr(points, '__len__'):
        indices = [ # And make sure to reverse to obtain row, column order
            tuple(reversed(tuple(
                map(int, ~transform * pair)))) for pair in points
        ]
    else:
        raise NotImplementedError('Argument "points" must be a sequence type')

    if hasattr(data, 'GetGeoTransform'):
        return sample(data.ReadAsArray(), indices)
    return sample(data, indices)


def xy_to_pixel(coords, gt, coords_srs_epsg = None, raster_srs_epsg = None):
    '''
    Translates a generic X-Y coordinate pair associated with a point into
    pixel coordinates, i.e., the row-column indices in a raster that contains
    that contains that point.

    Parameters
    ----------
    coords : tuple or list
        Coordinate pair, as a 2-element sequence
    gt : tuple
        GeoTransform for the associated raster
    coords_srs_epsg : int
        (Optional) EPSG code for the coordinate pair's SRS
    raster_srs_epsg : int
        (Optional) EPSG code for the associated raster

    Returns
    -------
    tuple
        A 2-element tuple; the row, column indices for the coordinate pair
    '''
    # Get an affine transformation to translate geographic coordinates to
    #   raster coordinates
    affine_transform = Affine.from_gdal(*gt)
    x, y = coords
    # Optionally, transform the geographic coordinates if needed
    if coords_srs_epsg is not None and raster_srs_epsg is not None:
        coords_srs = osr.SpatialReference()
        coords_srs.ImportFromEPSG(coords_srs_epsg)
        # Create a coordinate reference system transformation to translate
        #   latitude-longitude coordinates into EASE-Grid 2.0 coordinates
        if coords_srs.IsGeographic() == 1:
            coords = reversed(coords)
        x, y = crs_transform(coords_srs_epsg, raster_srs_epsg)\
            .TransformPoint(*coords, 0)[:-1] # Drop Z coordinate
    return tuple(map(int, reversed(~affine_transform * (x, y))))
